#!/usr/bin/python3
"""TEST using the FULL set of python-requirements: create the default example that all installations create and verify it thoroughly """
import logging, warnings, unittest, tempfile, os
from pathlib import Path
from pasta_eln.backend import Backend
from pasta_eln.AddOns.project_importCSV import main

class TestStringMethods(unittest.TestCase):
  """
  derived class for this test
  """
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.be = None
    self.dirName = ''

  def test_main(self):
    """
    main function
    """
    # initialization: create database, destroy on filesystem and database and then create new one
    warnings.filterwarnings('ignore', message='numpy.ufunc size changed')
    warnings.filterwarnings('ignore', message='invalid escape sequence')
    warnings.filterwarnings('ignore', category=ResourceWarning, module='PIL')
    warnings.filterwarnings('ignore', category=ImportWarning)
    logPath = Path.home()/'pastaELN.log'
    logging.basicConfig(filename=logPath, level=logging.INFO, format='%(asctime)s|%(levelname)s:%(message)s',
                        datefmt='%m-%d %H:%M:%S')   #This logging is always info, since for installation only
    for package in ['urllib3', 'requests', 'asyncio', 'PIL', 'matplotlib.font_manager']:
      logging.getLogger(package).setLevel(logging.WARNING)
    logging.info('Start test')

    # create .eln
    self.be = Backend('research')
    projID = self.be.output('x0').split('|')[-1].strip()
    self.be.changeHierarchy(projID)

    folderID = [i for i in self.be.outputHierarchy(True, True).split('\n') if 'This is an example task' in i][0]
    folderID = folderID.split('|')[-1].strip()
    self.be.changeHierarchy(folderID)
    main(self.be, '/'.join(self.be.hierStack), None, {'fileNames':['tests/inputSamples.csv']})

    self.verify()
    self.be.changeHierarchy(None)
    output = self.be.outputHierarchy(False)
    print(output)
    self.assertIn('sample A | sample', output, 'Sample A incorrect')
    self.assertIn('sample B | sample', output, 'Sample B incorrect')
    self.assertIn('sample C | sample', output, 'Sample C incorrect')
    print(f'{"*"*40}\nEND TEST 02 \n{"*"*40}')
    return


  def verify(self):
    #Verify DB
    output = self.be.checkDB(outputStyle='text')
    output = '\n'.join(output.split('\n')[8:])
    print(output)
    self.assertNotIn('**ERROR', output, 'Error in checkDB')
    self.assertLessEqual(len(output.split('\n')), 6, 'Check db should have less than 6 almost empty lines')
    return


  def tearDown(self):
    logging.info('End test')
    return

if __name__ == '__main__':
  unittest.main()



"""example addon: use a csv file to input data in rows

THIS IS A VERY ADVANCED ADDON TUTORIAL
This tutorial teaches
- how to group import metadata
"""
from pathlib import Path
import pandas as pd
from PySide6.QtWidgets import QFileDialog
from pasta_eln.guiStyle import showMessage
from pasta_eln.sqlite import MAIN_ORDER

# The following two variables are mandatory
description  = 'Import csv-data'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}

## Example csv file
# "type","name","comment",".chemistry","geometry.height","geometry.width"
# "sample","sample A","some text","A2B2C3",4,2
# "sample","sample B","other text","A2B2C4",4,4
# "sample","sample C","more text","A2B2C5",4,4

def main(backend, hierStack, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        backend (pasta backend): allow to extract data
        hierStack (list): node in hierarchy to start the creation
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    # Read csv file as a dataframe
    if 'fileNames' not in parameter:
        res = QFileDialog.getSaveFileName(widget,'Use this file for output', str(Path.home()))
        if res is None:
            return False
    else:
        res = parameter['fileNames']
    df = pd.read_csv(res[0])

    # verify the columns are correct
    colNames = list(df.columns)
    if 'type' not in colNames or 'name' not in colNames:
        showMessage(widget, 'Error', 'You have to have columns named "type" and "name"', 'Critical')
        return False
    if len(df['type'].unique()) > 1:
        showMessage(widget, 'Error', 'All items in the type column have to be the same', 'Critical')
        return False
    docType = df['type'].unique()[0]
    if docType not in backend.db.dataHierarchy('',''):
        showMessage(widget, 'Error', 'The type does not exist in PASTA database', 'Critical')
        return False
    # columns that are in the Pasta-ELN
    colPasta = [f'{i["class"]}.{i["name"]}' for i in backend.db.dataHierarchy(docType,'meta')]
    colPasta = [i[1:] if i[1:] in MAIN_ORDER+['tags','qrCodes'] else i for i in colPasta] + ['type']
    if set(colNames).difference(colPasta):
        showMessage(widget, 'Error', f'All columns must exist in the data schema. Offending: {set(colNames).difference(colPasta)}', 'Critical')
        return False

    # Move into that folder
    for i in hierStack.split('/'):
        backend.changeHierarchy(i)

    # Loop all rows
    for index, row in df.iterrows():
        data = row.to_dict()
        del data['type']
        backend.addData(docType, data)
    return True



""" Upload for Zenodo and Dataverse """
import json
import tempfile
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any
import qtawesome as qta
from PySide6.QtCore import Qt  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import QCheckBox, QDialog, QLabel, QLineEdit, QVBoxLayout  # pylint: disable=no-name-in-module
from ...fixedStringsJson import CONF_FILE_NAME
from ...guiCommunicate import Communicate
from ...guiStyle import Label, TextButton, showMessage, widgetAndLayout, widgetAndLayoutGrid
from ...inputOutput import exportELN
from .dataverse import DataverseClient
from .zenodo import ZenodoClient


class UploadGUI(QDialog):
  """ Upload for Zenodo and Dataverse """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      backend (Backend): PASTA-ELN backend
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm    = comm
    self.setWindowTitle('Upload project to repository')

    # GUI elements
    mainL = QVBoxLayout(self)
    Label('Upload to a repository', 'h1', mainL)
    _, center = widgetAndLayout('H', mainL, spacing='l', bottom='l', top='m')

    if not self.comm.projectID:
      return
    docProject = self.comm.backend.db.getDoc(self.comm.projectID)
    repositories = self.comm.backend.configuration['repositories']
    leftSideW, leftSide = widgetAndLayoutGrid(center, spacing='m', right='l')
    leftSideW.setStyleSheet('border-right: 2px solid black;')
    leftSide.setAlignment(Qt.AlignTop)                                            # type: ignore[attr-defined]
    leftSide.addWidget(Label('Metadata','h2'), 0, 0)
    leftSide.addWidget(QLabel('Title'), 1, 0)
    self.leTitle = QLineEdit(docProject['name'])
    self.leTitle.setMinimumWidth(400)
    leftSide.addWidget(self.leTitle, 1, 1)
    leftSide.addWidget(QLabel('Description'), 2, 0)
    self.leDescription = QLineEdit(docProject['comment'])
    leftSide.addWidget(self.leDescription, 2, 1)
    leftSide.addWidget(QLabel('Keywords'), 3, 0)
    self.leKeywords = QLineEdit(', '.join(docProject['tags']))
    leftSide.addWidget(self.leKeywords, 3, 1)
    leftSide.addWidget(QLabel('Category/Community'), 4, 0)
    self.leCategory = QLineEdit(repositories.get('category',''))
    leftSide.addWidget(self.leCategory, 4, 1)
    leftSide.addWidget(QLabel('Additional'), 5, 0)
    self.leAdditional = QLineEdit(json.dumps(repositories.get('additional',{})))
    leftSide.addWidget(self.leAdditional, 5, 1)

    self.allDocTypes = self.comm.backend.db.dataHierarchy('', 'title')
    projectString = ', '.join(i[1] for i in self.allDocTypes if i[0].startswith('x'))
    _, rightSide = widgetAndLayout('V', center, spacing='m', right='l')
    rightSide.setAlignment(Qt.AlignTop)                                           # type: ignore[attr-defined]
    Label('Include data types','h2', rightSide)
    self.allCheckboxes = [QCheckBox(projectString, self)]
    self.allCheckboxes[0].setChecked(True)
    self.allCheckboxes[0].setDisabled(True)
    rightSide.addWidget(self.allCheckboxes[0])
    for i in self.allDocTypes:
      if not i[0].startswith('x') and '/' not in i[0]:
        checkbox = QCheckBox(i[1], self)
        checkbox.setChecked(True)
        self.allCheckboxes.append(checkbox)
        rightSide.addWidget(checkbox)

    #final button box
    _, buttonLineL = widgetAndLayout('H', mainL, 'm')
    buttonLineL.addStretch(1)
    if 'zenodo' in repositories and 'url' in repositories['zenodo']:
      TextButton('Upload to Zenodo',    self, [Command.UPLOAD, True],  buttonLineL)
    if 'dataverse' in repositories and 'url' in repositories['dataverse']:
      TextButton('Upload to Dataverse', self, [Command.UPLOAD, False], buttonLineL)
    TextButton('Cancel',              self, [Command.CANCEL], buttonLineL, 'Discard changes')


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.CANCEL:
      self.reject()
    elif command[0] is Command.UPLOAD:
      # collect docTypes and create .eln
      docTypes = [i.text() for i in self.allCheckboxes if i.isChecked() and ', ' not in i.text()]
      tempELN = str(Path(tempfile.gettempdir())/'export.eln')
      res0 = exportELN(self.comm.backend, [self.comm.projectID], tempELN, docTypes)
      print('Export eln',res0)

      # collect metadata and save parts of it
      metadata = {'title': self.leTitle.text(),
                  'description': self.leDescription.text(),
                  'keywords': [i.strip() for i in self.leKeywords.text().split(',')],
                  'category': self.leCategory.text(),
                  'additional': json.loads(self.leAdditional.text()),
                  'author': self.comm.backend.configuration['authors'][0]}
      repositories = self.comm.backend.configuration['repositories']
      repositories['category'] = self.leCategory.text()
      repositories['additional'] = json.loads(self.leAdditional.text())
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.backend.configuration,indent=2))
      # convert to repository specific and upload to repository
      if command[1]: #Zenodo
        clientZ = ZenodoClient(repositories['zenodo']['url'], repositories['zenodo']['key'])
        metadataZ = clientZ.prepareMetadata(metadata)
        res = clientZ.uploadRepository(metadataZ, tempELN)
      else:          #Dataverse
        clientD = DataverseClient(repositories['dataverse']['url'], repositories['dataverse']['key'],
                                  repositories['dataverse']['dataverse'])
        metadataD = clientD.prepareMetadata(metadata)
        res = clientD.uploadRepository(metadataD, tempELN)
      showMessage(self, 'Success' if res[0] else 'Error', res[1])
      # update project with upload details
      if res[0]:
        docProject = self.comm.backend.db.getDoc(self.comm.projectID)
        docProject['.repository_upload'] = f'{datetime.now().strftime("%Y-%m-%d")} {res[1]}'
        docProject['branch'] = docProject['branch'][0] | {'op':'u'}
        self.comm.backend.db.updateDoc(docProject, self.comm.projectID)
      else:
        showMessage(self, 'Error', res[1])
      self.accept()
    else:
      print('Got some button, without definition', command)
    return


  def changeButtonOnTest(self, success:bool, button:TextButton, message:str='') -> None:
    """ Helper function to change buttons upon success/failure

    Args:
      success (bool): change to successful state
      button (TextBotton): button to change
      message (str): optional message on failure
    """
    if success:
      button.setStyleSheet('background: #00FF00')
      button.setText('')
      button.setIcon(qta.icon('fa5s.check-square', scale_factor=1))
    else:
      if message:
        showMessage(self, 'Error', message)
      button.setStyleSheet('background: #FF0000')
      button.setText('')
      button.setIcon(qta.icon('fa5.times-circle', scale_factor=1))
    return


class Command(Enum):
  """ Commands used in this file """
  UPLOAD = 1
  CANCEL = 2



"""extract data from vendor
- default png file

THIS IS THE ADVANCED EXTRACTOR TUTORIAL, READ THE BASIC TUTORIAL (_csv.py) FIRST.
This tutorial teaches
- how to use style recipes with choices
  - you can specify sub-choices with the dict, as shown.
  - what you from the calling function is a string
  - for getting the value from the string, you might have to use split / strip, as shown
  - you can use as many dictionary entries as you want
  - just make sure dictionary entries are correct
  - please use this way of writing code and the extractors should work
  - for debugging: print / debuggers are working as normal
- how to create metadata data-scientifically using a list of dictionaries
  - here you can also use your language specific symbols
  - here you can also add uncertainty. If you need your uncertainty in a separate section, do so by creating a separate entry
- how to create images and scale them down to save space in the meta-data-base
- since this is a tutorial, it might not make scientific sense. The goal is teaching!
"""
import base64
from io import BytesIO
import numpy as np
from PIL import Image, ImageFilter


def use(filePath, style={'main':''}, saveFileName=None):
  """
  Args:
    filePath (Path): full path file name
    style    (dict): supplied to guide the display / extraction style = recipe
                     main is / separated hierarchical elements parent->child
                     can contain more elements
    saveFileName (string): if given, save the image to this file-name

  Returns:
    dict: containing image, metaVendor, metaUser, style
  """
  # THIS IS THE IMPORTANT PART OF THE EXTRACTOR
  image = Image.open(filePath)
  metaVendor = image.info
  toDelete = [key for key, value in metaVendor.items() if not isinstance(value, (str, int, float, list))]
  for key in toDelete:
    del metaVendor[key]
  if style['main'] == 'measurement/image/crop':    #: Crop 3/4 of the image
    imgArr = np.array(image)
    newHeight = int(imgArr.shape[0]/2)
    newWidth  = int(imgArr.shape[1]/2)
    imgArr = imgArr[:newHeight, :newWidth, :]
    image = Image.fromarray(imgArr)
  elif style['main'] == 'measurement/image/gaussian':    #: Gaussian blur {'blur radius':['r = 1','r = 5','r = 10','r = 100']}
    radius = int(style.get('blur radius','r = 1').split('=')[-1].strip())
    image = image.filter(ImageFilter.GaussianBlur(radius = radius))
  elif True or style['main'] == 'measurement/image': #: Default | uncropped
    style['main'] = 'measurement/image'
  metaUser   = [{'key':'imageWidth',  'value':image.size[0], 'unit':'mm', 'label':'Largeur de l`image', 'PURL':'http://purl.allotrope.org/ontologies/result#AFR_0002468'},
                {'key':'imageHeight', 'value':f'{image.size[1]}+/- 3', 'unit':'mm', 'label':'Höhe des Bildes','PURL':'http://purl.allotrope.org/ontologies/result#AFR_0002467'}
               ]
  # save to file: this is the high quality image
  if saveFileName is not None:
    image.save(saveFileName)

  # convert PIL image to base64: here one should downscale
  maxSize = 400
  if max(image.size)>maxSize:
    scale = max(image.size)/maxSize
    newSize = (int(image.size[0]/scale), int(image.size[1]/scale))
    image = image.resize(newSize)
  figfile = BytesIO()
  image.save(figfile, format='PNG')
  imageB64 = base64.b64encode(figfile.getvalue()).decode()
  imageB64 = f"data:image/png;base64,{imageB64}"

  # return everything
  return {'image':imageB64, 'style':style, 'metaVendor':metaVendor, 'metaUser':metaUser}

  #other datatypes could follow here if statements are used
  #...

  #final return if nothing successful
  #return {'style': '-', 'metaVendor':{}, 'metaUser':{}}



""" Interactions with Zenodo repository """
from datetime import datetime
from typing import Any
import requests
from .repository import RepositoryClient


class ZenodoClient(RepositoryClient):
  """ Interactions with Zenodo repository """
  def __init__(self, server_url: str, api_token: str) -> None:
    """
    Initializes the client.

    Args:
        server_url (str): The URL of the server.
        api_token (str): The API token for authentication.
    """
    super().__init__(server_url, api_token)
    self.headers1 = {'Content-Type': 'application/json', 'Authorization': f"Bearer {api_token}"}
    self.headers2 = {'Authorization': f"Bearer {api_token}"}

  def checkServer(self) -> tuple[bool, str]:
    """ VOID TEST SINCE ZENODO DOES NOT HAVE A SERVER TEST

    Checks if the data-verse server is reachable

    Returns (tuple(bool, Any)):
      A tuple of (success, a message) is returned
    """
    return True, 'VOID TEST'
    # resp = requests.get(f"{self.server_url}/api/info/version", headers=self.headers1)
    # success = (resp.status_code == 200 and resp.json().get('data').get('version') is not None)
    # return (success, 'Dataverse is reachable') \
    #   if success \
    #   else (success, f"Cannot reach server: {self.server_url}, Status: {resp.status_code}, json: {resp.json()}")


  def checkAPIKey(self) -> bool:
    """
    Checks if the given API token is valid.

    Explanation:
        This method checks if the provided API token is valid by making a request to the server.
        It logs the server URL and sends a GET request to the token endpoint with the API token.
        It returns True if the response is successful and the status code is not 401, 403, or 500.

    Args:
        self: The instance of the class.

    Returns:
        bool: True if the API token is valid, False otherwise.
    """
    resp = requests.get(f"{self.server_url}", headers=self.headers1, timeout=10)
    return (bool(resp) and resp.status_code is not None
            and resp.status_code not in [401, 403, 500])


  def uploadRepository(self, metadata:dict[str,Any], file_path:str) -> tuple[bool, str]:
    """
    Uploads a file and metadata to become a dataset.

    Args:
      metadata (dict): metadata to this file according to Zenodo standard
      file_path (str): The absolute path to the file to be uploaded.

    Returns:
      tuple: success of function, message
    """
    server_url = f"{self.server_url}/api/deposit/depositions"
    # Define the API URLs and headers based on the repository kind
    # Step 1: Create the deposition with metadata
    resp = requests.post(server_url, json=metadata, headers=self.headers1, timeout=10)
    if resp.status_code != 201:
      print('**ERROR** creating deposition/dataset:', resp.json(), resp.status_code, resp.text)
      return False, 'Error creating the dataset'
    deposition = resp.json()
    persistentID = deposition['id']
    # print(f"Deposition created: {persistentID}")

    # Define the API URLs and headers based on the repository kind
    with open(file_path, 'rb') as f:
      files = {'file': f}
    file_upload_url = f"{server_url}/{persistentID}/files"
    publish_url = f"{server_url}/{persistentID}/actions/publish"

    # Step 2: Upload a file
    resp = requests.post(file_upload_url, files=files, headers=self.headers2, timeout=10)
    if resp.status_code != 201:
      print('**ERROR** uploading file:', resp.json())
      return False, 'Error uploading the file'
    # print("File uploaded successfully:")

    # Step 3: Publish the deposition
    resp = requests.post(publish_url, headers=self.headers1, timeout=10)
    if resp.status_code != 202:
      print('**ERROR** publishing:', resp.json())
      return False, 'Error publishing the dataset'
    return True, f'Published: {resp.json()["doi"]}, {resp.json()["doi_url"]}'


  def prepareMetadata(self, metadata:dict[str,Any]) -> dict[str,Any]:
    """
    Prepares the metadata for uploading.

    Args:
        metadata (dict): The metadata to be prepared.

    Returns:
        dict: The prepared metadata.
    """
    author = metadata['author']
    metadataZenodo = {
        'title': metadata['title'],
        'upload_type': 'dataset',
        'description': metadata['description'],
        'creators': [
            {'name': f"{author['last']}, {author['first']}",
             'affiliation': author['organizations'][0]['organization'],
             'orcid': author['orcid'],
             'email': author['email']},
        ],
        'keywords': metadata['keywords'],
        'communities': [
            {'identifier': metadata['category']},
        ],
        'publication_date': datetime.now().strftime('%Y-%m-%d'),
        'access_right': 'open',
        'license': 'CC-BY-4.0'
      }
    return {'metadata': metadata['additional']|metadataZenodo}



""" ListFeeDelegate module used for the table views """
from PySide6.QtCore import QModelIndex, QPersistentModelIndex
from PySide6.QtWidgets import QLineEdit, QStyledItemDelegate, QStyleOptionViewItem, QWidget


class ListFreeDelegate(QStyledItemDelegate):
  """
  Delegate for creating the line edit with lookup icon for the iri column in data hierarchy editor tables
  """
  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index:  QModelIndex | QPersistentModelIndex) -> QWidget:
    """
    Creates the line edit

    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: QLineEdit widget
    """
    #remove listItem data
    indexListItem = index.model().index(index.row(), 4)
    index.model().setData(indexListItem, '')
    #normal stuff
    return QLineEdit(parent)



""" Edit properties of a docType """
import random
import string
from typing import Callable, Optional
import qtawesome as qta
from PySide6.QtGui import QRegularExpressionValidator  # pylint: disable=no-name-in-module
from PySide6.QtWidgets import (QComboBox, QDialog, QDialogButtonBox, QLabel,  # pylint: disable=no-name-in-module
                               QLineEdit, QVBoxLayout)
from ...fixedStringsJson import allIcons
from ...guiCommunicate import Communicate
from ...guiStyle import TextButton, showMessage, widgetAndLayout, widgetAndLayoutForm


class DocTypeEditor(QDialog):
  """ Edit properties of a docType """
  def __init__(self, comm:Communicate, docType:str, callback:Optional[Callable[[str],None]]=None):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
      doc (dict):  document to change / create
      callback (function): callback to allow sending the new doc-type
    """
    super().__init__()
    self.comm = comm
    self.db   = self.comm.backend.db
    self.docType = docType
    self.callback = callback
    mainL = QVBoxLayout(self)
    _, mainForm = widgetAndLayoutForm(mainL)
    self.setWindowTitle('Edit docType properties')

    initialData = self.comm.backend.db.dataHierarchy(docType,'*') if docType else ['']*6
    self.row1 = QLineEdit(initialData[0])
    self.row1.setToolTip('Name of document type: lower case letters only. Must be unique')
    if docType:
      self.row1.setDisabled(True)
    else:
      self.row1.setValidator(QRegularExpressionValidator(r'(^[a-wyz][\w\/]{3,}$)'))
    mainForm.addRow(QLabel('DocType '), self.row1)

    self.row2 = QLineEdit(initialData[2])
    self.row2.setToolTip('Label that the user reads: it is suggested to start with upper case and end with s as in "Samples"')
    mainForm.addRow(QLabel('Label '), self.row2)

    row3W, row3L = widgetAndLayout('H', None, 's')
    # Label('type:', 'h2', row3L)
    # self.comboType = QComboBox()
    # self.comboType.addItems(set(i.split('.')[0] for i in allIcons))
    # row3L.addWidget(self.comboType)
    # Label('icon:', 'h2', row3L)
    self.comboIcon = QComboBox()
    self.comboIcon.addItem('')
    for icon in allIcons:
      self.comboIcon.addItem(qta.icon(icon, scale_factor=1.5), icon)
    self.comboIcon.setCurrentText(initialData[3])
    row3L.addWidget(self.comboIcon)
    if docType.startswith('x'):
      self.comboIcon.setDisabled(True)
    mainForm.addRow(QLabel('Icon '), row3W)

    self.row4 = QLineEdit(initialData[4])
    self.row4.setToolTip('One letter shortcut used with Ctrl-. Must be unique.')
    if docType.startswith('x'):
      self.row4.setDisabled(True)
    else:
      data = [i[1] for i in self.comm.backend.db.dataHierarchy('','shortcut') if not i[0].startswith('x')]
      dataSet = set(string.ascii_lowercase).difference(data)
      self.row4.setValidator(QRegularExpressionValidator(f'^[{"".join(dataSet)}]$'))
    mainForm.addRow(QLabel('Shortcut '), self.row4)

    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    else:
      label    = self.row2.text()
      if not label or label in [v1 for v0,v1 in self.db.dataHierarchy('','title')]:
        label = 'Random_'+''.join(random.choices(string.ascii_uppercase + string.digits, k=6))
      icon     = '' if self.docType.startswith('x') else self.comboIcon.currentText()
      shortcut = '' if self.docType.startswith('x') else self.row4.text()
      if self.docType:   # update existing docType
        self.comm.backend.db.cursor.execute(f"UPDATE docTypes SET title='{label}', shortcut='{shortcut}', "\
                                            f"icon='{icon}' WHERE docType = '{self.docType}'")
      else:              # create new docType, with default schema entries
        docType = self.row1.text()
        if not docType or docType in self.comm.backend.db.dataHierarchy('',''):
          showMessage(self, 'Error', 'DocType name is not valid or already exists', 'Warning')
          return
        self.comm.backend.db.cursor.execute('INSERT INTO docTypes VALUES (?, ?, ?, ?, ?, ?)',
                                            [docType, '', label, icon, shortcut, ''])
        self.comm.backend.db.cursor.execute('INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                            [docType, '', '0', 'name', '', '', ''])
        self.comm.backend.db.cursor.execute('INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                            [docType, '', '1', 'tags', '', '', ''])
        self.comm.backend.db.cursor.execute('INSERT INTO docTypeSchema VALUES (?, ?, ?, ?, ?, ?, ?)',
                                            [docType, '', '2', 'comment', '', '', ''])
        if self.callback is not None:
          self.callback(label)
          #TODO: callback has to be separated: changeDocType in editor.py + flag for soft-restart comm signal
          #    many places should use this signal
      self.comm.backend.db.connection.commit()
      self.accept()
    return



# pylint: skip-file
# mypy: ignore-errors
from __future__ import annotations
import base64
import enum
import hashlib
import os
import secrets
import time
from dataclasses import dataclass
from pathlib import Path
from typing import BinaryIO, Optional, Union
from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.kdf import scrypt
from .exceptions import Error, ParseError, VerifyError
from .helpers import Reader, check_comment, read_data

ALG_LEN = 2
KDF_PARAM_LEN = 8
KEY_ID_LEN = 8
KEY_LEN = 32
SALT_LEN = 32
CHECKSUM_LEN = 32
SIG_LEN = 64

KEYNUM_PK_LEN = KEY_ID_LEN + KEY_LEN
KEYNUM_SK_LEN = KEY_ID_LEN + (KEY_LEN << 1) + CHECKSUM_LEN

OPSLIMIT = 1_048_576
MEMLIMIT = 33_554_432
MEMLIMIT_MAX = 1_073_741_824
N_LOG2_MAX = 20

SIG_EXT = 'minisig'
BYTE_ORDER = 'little'
DEFAULT_SK_PATH = '~/.minisign/minisign.key'

UNTRUSTED_COMMENT_PREFIX = 'untrusted comment: '
TRUSTED_COMMENT_PREFIX = 'trusted comment: '
TRUSTED_COMMENT_PREFIX_LEN = len(TRUSTED_COMMENT_PREFIX)


@enum.unique
class SignatureAlgorithm(bytes, enum.Enum):
    PURE_ED_DSA = bytes([0x45, 0x64])
    PREHASHED_ED_DSA = bytes([0x45, 0x44])


@enum.unique
class KDFAlgorithm(bytes, enum.Enum):
    SCRYPT = bytes([0x53, 0x63])


@enum.unique
class CksumAlgorithm(bytes, enum.Enum):
    BLAKE2b = bytes([0x42, 0x32])


@dataclass(frozen=True)
class Signature:
    _untrusted_comment: str
    _signature_algorithm: SignatureAlgorithm
    _key_id: bytes
    _signature: bytes
    _trusted_comment: str
    _global_signature: bytes

    @classmethod
    def from_bytes(cls, data: bytes) -> Signature:
        lines = data.splitlines()
        if len(lines) < 4:
            raise ParseError('incomplete encoded signature')
        glob_sig = base64.standard_b64decode(lines[3])
        if len(glob_sig) != SIG_LEN:
            raise ParseError('invalid encoded signature')
        buf = Reader(base64.standard_b64decode(lines[1]))
        return cls(
            _untrusted_comment=lines[0].decode(),
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _key_id=buf.read(KEY_ID_LEN),
            _signature=buf.read(SIG_LEN),
            _trusted_comment=lines[2].decode(),
            _global_signature=glob_sig,
        )

    @classmethod
    def from_file(cls, path: str | os.PathLike) -> Signature:
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @property
    def untrusted_comment(self) -> str:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: str):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    @property
    def trusted_comment(self) -> str:
        return self._trusted_comment[TRUSTED_COMMENT_PREFIX_LEN:]

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            self._untrusted_comment.encode(),
            base64.standard_b64encode(
                self._signature_algorithm.value +
                self._key_id +
                self._signature
            ),
            self._trusted_comment.encode(),
            base64.standard_b64encode(self._global_signature),
        ))

    def _is_prehashed(self) -> bool:
        return self._signature_algorithm == SignatureAlgorithm.PREHASHED_ED_DSA


@dataclass(frozen=True)
class KeynumPK:
    key_id: bytes
    public_key: bytes

    @classmethod
    def from_bytes(cls, data: bytes | Reader) -> KeynumPK:
        assert len(data) == KEYNUM_PK_LEN
        if isinstance(data, bytes):
            data = Reader(data)
        return cls(
            key_id=data.read(KEY_ID_LEN),
            public_key=data.read(KEY_LEN),
        )

    def __bytes__(self) -> bytes:
        return self.key_id + self.public_key


@dataclass(frozen=True)
class PublicKey:
    _untrusted_comment: str | None
    _signature_algorithm: SignatureAlgorithm
    _keynum_pk: KeynumPK

    @classmethod
    def from_base64(cls, s: bytes | str) -> PublicKey:
        buf = Reader(base64.standard_b64decode(s))
        return cls(
            _untrusted_comment=None,
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _keynum_pk=KeynumPK.from_bytes(buf),
        )

    @classmethod
    def from_bytes(cls, data: bytes) -> PublicKey:
        lines = data.splitlines()
        if len(lines) < 2:
            raise ParseError('incomplete encoded public key')
        pk = cls.from_base64(lines[1])
        pk.set_untrusted_comment(lines[0].decode())
        return pk

    @classmethod
    def from_file(cls, path: str | os.PathLike) -> PublicKey:
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @classmethod
    def from_secret_key(cls, secret_key: SecretKey) -> PublicKey:
        key_id = bytes(secret_key._keynum_sk.key_id)
        return cls(
            _untrusted_comment=f'{UNTRUSTED_COMMENT_PREFIX}'
                               f'minisign public key '
                               f'{key_id.hex().upper()}',
            _signature_algorithm=secret_key._signature_algorithm,
            _keynum_pk=KeynumPK(
                key_id=key_id,
                public_key=bytes(secret_key._keynum_sk.public_key),
            ),
        )

    @property
    def untrusted_comment(self) -> str | None:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: str | None):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    def verify(self, data: bytes | BinaryIO, signature: Signature):
        if self._keynum_pk.key_id != signature._key_id:
            raise VerifyError('incompatible key identifiers')
        if not signature._trusted_comment.startswith(TRUSTED_COMMENT_PREFIX):
            raise VerifyError('unexpected format for the trusted comment')
        pk = ed25519.Ed25519PublicKey.from_public_bytes(
            self._keynum_pk.public_key)
        try:
            pk.verify(
                signature._signature,
                read_data(data, signature._is_prehashed()),
            )
            pk.verify(
                signature._global_signature,
                signature._signature + signature.trusted_comment.encode(),
            )
        except InvalidSignature as err:
            raise VerifyError(err) from err

    def verify_file(
        self,
        path: str | os.PathLike,
        signature: Signature | None = None,
    ):
        if signature is None:
            signature = Signature.from_file(f'{path}.{SIG_EXT}')
        with open(path, 'rb') as f:
            self.verify(f, signature)

    def to_base64(self) -> bytes:
        return base64.standard_b64encode(
            self._signature_algorithm.value +
            bytes(self._keynum_pk)
        )

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            (
                f'{UNTRUSTED_COMMENT_PREFIX}minisign public key '
                f'{self._keynum_pk.key_id.hex().upper()}'
                if self._untrusted_comment is None else
                self._untrusted_comment
            ).encode(),
            self.to_base64(),
        ))


@dataclass(frozen=True, repr=False)
class KeynumSK:
    key_id: bytearray
    secret_key: bytearray
    public_key: bytearray
    checksum: bytearray

    @classmethod
    def from_bytes(cls, data: bytes | Reader) -> KeynumSK:
        assert len(data) == KEYNUM_SK_LEN
        if isinstance(data, bytes):
            data = Reader(data)
        return cls(
            key_id=bytearray(data.read(KEY_ID_LEN)),
            secret_key=bytearray(data.read(KEY_LEN)),
            public_key=bytearray(data.read(KEY_LEN)),
            checksum=bytearray(data.read(CHECKSUM_LEN)),
        )

    def xor(self, key: bytes):
        assert len(key) == KEYNUM_SK_LEN
        buf = Reader(key)
        for (l, size) in (
            (self.key_id, KEY_ID_LEN),
            (self.secret_key, KEY_LEN),
            (self.public_key, KEY_LEN),
            (self.checksum, CHECKSUM_LEN),
        ):
            for idx, (v1, v2) in enumerate(zip(l[:], buf.read(size))):
                l[idx] = v1 ^ v2

    def __bytes__(self) -> bytes:
        return (
            bytes(self.key_id) +
            bytes(self.secret_key) +
            bytes(self.public_key) +
            bytes(self.checksum)
        )


@dataclass(frozen=True, repr=False)
class SecretKey:
    _untrusted_comment: str
    _signature_algorithm: SignatureAlgorithm
    _kdf_algorithm: KDFAlgorithm
    _cksum_algorithm: CksumAlgorithm
    _kdf_salt: bytes
    _kdf_opslimit: int
    _kdf_memlimit: int
    _keynum_sk: KeynumSK

    @classmethod
    def from_bytes(cls, data: bytes) -> SecretKey:
        lines = data.splitlines()
        if len(lines) < 2:
            raise ParseError('incomplete encoded secret key')
        buf = Reader(base64.standard_b64decode(lines[1]))
        return cls(
            _untrusted_comment=lines[0].decode(),
            _signature_algorithm=SignatureAlgorithm(buf.read(ALG_LEN)),
            _kdf_algorithm=KDFAlgorithm(buf.read(ALG_LEN)),
            _cksum_algorithm=CksumAlgorithm(buf.read(ALG_LEN)),
            _kdf_salt=buf.read(SALT_LEN),
            _kdf_opslimit=int.from_bytes(buf.read(KDF_PARAM_LEN), BYTE_ORDER),
            _kdf_memlimit=int.from_bytes(buf.read(KDF_PARAM_LEN), BYTE_ORDER),
            _keynum_sk=KeynumSK.from_bytes(buf),
        )

    @classmethod
    def from_file(
        cls,
        path: str | os.PathLike | None = None,
    ) -> SecretKey:
        if path is None:
            path = Path(DEFAULT_SK_PATH).expanduser().resolve(strict=True)
        with open(path, 'rb') as f:
            return cls.from_bytes(f.read())

    @property
    def untrusted_comment(self) -> str:
        return self._untrusted_comment

    def set_untrusted_comment(self, value: str):
        check_comment(value)
        self.__dict__['_untrusted_comment'] = value

    def get_public_key(self) -> PublicKey:
        return PublicKey.from_secret_key(self)

    def decrypt(self, password: str):
        self._crypt(password)
        if self._calc_checksum() != bytes(self._keynum_sk.checksum):
            raise Error('wrong password for that key')

    def encrypt(self, password: str):
        self._crypt(password)

    def _crypt(self, password: str):
        if self._kdf_memlimit > MEMLIMIT_MAX:
            raise Error('memlimit too high')
        opslimit = max(32768, self._kdf_opslimit)
        n_log2 = 1
        r = 8
        p = 0
        if opslimit < self._kdf_memlimit // 32:
            maxn = opslimit // (r * 4)
            p = 1
        else:
            maxn = self._kdf_memlimit // (r * 128)
        while n_log2 < 63 and not 1 << n_log2 > maxn // 2:
            n_log2 += 1
        if not p:
            p = min(0x3fffffff, (opslimit // 4) // (1 << n_log2)) // r
        if n_log2 > N_LOG2_MAX:
            raise Error('n_log2 too high')
        self._keynum_sk.xor(scrypt.Scrypt(
            salt=self._kdf_salt,
            length=KEYNUM_SK_LEN,
            n=1 << n_log2,
            r=r,
            p=p,
        ).derive(password.encode()))

    def sign(
        self,
        data: bytes | BinaryIO,
        *,
        prehash: bool = True,
        untrusted_comment: str | None = None,
        trusted_comment: str | None = None,
    ) -> Signature:
        untrusted_comment = (
            f'{UNTRUSTED_COMMENT_PREFIX}minisign signature '
            f'{self._keynum_sk.key_id.hex().upper()}'
            if untrusted_comment is None else
            untrusted_comment
        )
        check_comment(untrusted_comment)
        trusted_comment = (
            f'timestamp:{int(time.time())}'
            if trusted_comment is None else
            trusted_comment
        )
        check_comment(trusted_comment)
        pk = ed25519.Ed25519PrivateKey.from_private_bytes(
            self._keynum_sk.secret_key)
        sig_sig = pk.sign(read_data(data, prehash))
        return Signature(
            _untrusted_comment=untrusted_comment,
            _signature_algorithm=(
                SignatureAlgorithm.PREHASHED_ED_DSA
                if prehash else
                SignatureAlgorithm.PURE_ED_DSA
            ),
            _key_id=self._keynum_sk.key_id,
            _signature=sig_sig,
            _trusted_comment=f'{TRUSTED_COMMENT_PREFIX}{trusted_comment}',
            _global_signature=pk.sign(sig_sig + trusted_comment.encode()),
        )

    def sign_file(
        self,
        path: str | os.PathLike,
        *,
        prehash: bool = False,
        untrusted_comment: str | None = None,
        trusted_comment: str | None = None,
        drop_signature: bool = False,
    ) -> Signature:
        with open(path, 'rb') as f:
            sig = self.sign(
                f,
                prehash=prehash,
                untrusted_comment=untrusted_comment,
                trusted_comment=trusted_comment,
            )
        if drop_signature:
            with open(f'{path}.{SIG_EXT}', 'wb') as f1:
                f1.write(bytes(sig))
                f1.write(b'\n')
        return sig

    def _calc_checksum(self) -> bytes:
        hasher = hashlib.blake2b(digest_size=CHECKSUM_LEN)
        hasher.update(self._signature_algorithm.value)
        hasher.update(self._keynum_sk.key_id)
        hasher.update(self._keynum_sk.secret_key)
        hasher.update(self._keynum_sk.public_key)
        return hasher.digest()

    def _update_checksum(self):
        self._keynum_sk.checksum[:] = self._calc_checksum()

    def __bytes__(self) -> bytes:
        return b'\n'.join((
            self._untrusted_comment.encode(),
            base64.standard_b64encode(
                self._signature_algorithm.value +
                self._kdf_algorithm.value +
                self._cksum_algorithm.value +
                self._kdf_salt +
                self._kdf_opslimit.to_bytes(KDF_PARAM_LEN, BYTE_ORDER) +
                self._kdf_memlimit.to_bytes(KDF_PARAM_LEN, BYTE_ORDER) +
                bytes(self._keynum_sk)
            ),
        ))


@dataclass(frozen=True, repr=False)
class KeyPair:
    secret_key: SecretKey
    public_key: PublicKey

    @classmethod
    def generate(cls) -> KeyPair:
        private_key = ed25519.Ed25519PrivateKey.generate()
        key_id = secrets.token_bytes(KEY_ID_LEN)
        sk = SecretKey(
            _untrusted_comment=f'{UNTRUSTED_COMMENT_PREFIX}'
                               f'minisign secret key '
                               f'{key_id.hex().upper()}',
            _signature_algorithm=SignatureAlgorithm.PURE_ED_DSA,
            _kdf_algorithm=KDFAlgorithm.SCRYPT,
            _cksum_algorithm=CksumAlgorithm.BLAKE2b,
            _kdf_salt=secrets.token_bytes(SALT_LEN),
            _kdf_opslimit=OPSLIMIT,
            _kdf_memlimit=MEMLIMIT,
            _keynum_sk=KeynumSK(
                key_id=bytearray(key_id),
                secret_key=bytearray(private_key.private_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PrivateFormat.Raw,
                    encryption_algorithm=serialization.NoEncryption(),
                )),
                public_key=bytearray(private_key.public_key().public_bytes(
                    encoding=serialization.Encoding.Raw,
                    format=serialization.PublicFormat.Raw,
                )),
                checksum=bytearray(CHECKSUM_LEN),
            ),
        )
        sk._update_checksum()
        return cls(secret_key=sk, public_key=PublicKey.from_secret_key(sk))



""" ListItemDelegate module used for the table views """
from PySide6.QtCore import QAbstractItemModel, QModelIndex, QPersistentModelIndex, Qt
from PySide6.QtWidgets import QComboBox, QStyledItemDelegate, QStyleOptionViewItem, QWidget


class ListItemDelegate(QStyledItemDelegate):
  """
  Delegate for creating the line edit with lookup icon for the iri column in data hierarchy editor tables
  """
  def __init__(self, docTypesLabels:list[tuple[str,str]]):
    """
    Args:
      docTypeLabels (list): list of docTypes and docLabels
    """
    super().__init__()
    self.docTypesLabels = docTypesLabels


  def createEditor(self,
                   parent: QWidget,
                   option: QStyleOptionViewItem,
                   index:  QModelIndex | QPersistentModelIndex) -> QWidget:
    """
    Disable the editor for the whole column by simply returning None
    Args:
      parent (QWidget): Parent table view.
      option (QStyleOptionViewItem): Style option for the cell represented by index.
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.

    Returns: QLineEdit widget
    """
    # skip last row
    indexName = index.model().index(index.row(), 0)
    if not indexName.data():
      return None                                                                 # type: ignore[return-value]
    # remove listItem data
    indexListFree = index.model().index(index.row(), 5)
    index.model().setData(indexListFree, '')
    # do conventional step
    comboBox = QComboBox(parent)
    _ = [comboBox.addItem(text, data) for (data, text) in [('','')]+self.docTypesLabels]
    return comboBox

  def setEditorData(self,
                    editor: QWidget,
                    index: QModelIndex | QPersistentModelIndex) -> None:
    """
    Set the data in the editor

    Args:
      editor (QWidget): widget to set data in
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    current_text = index.data(Qt. EditRole)                                       # type: ignore[attr-defined]
    editor.setCurrentText(current_text)                                           # type: ignore[attr-defined]
    return


  def setModelData(self,
                   editor: QWidget,
                   model: QAbstractItemModel,
                   index: QModelIndex | QPersistentModelIndex) -> None:
    """
    Change data in model accordingly

    Args:
      editor (QWidget): widget to set data in
      model (QAbstractItemModel): model to change
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    model.setData(index, editor.currentText(), Qt.EditRole)                       # type: ignore[attr-defined]
    return


  def updateEditorGeometry(self,
                           editor: QWidget,
                           option: QStyleOptionViewItem,
                           index: QModelIndex | QPersistentModelIndex) -> None:
    """
    Change size of field

    Args:
      editor (QWidget): widget to set data in
      option (QStyledOptionViewItem): information on the geometry, same as paint
      index (Union[QModelIndex, QPersistentModelIndex]): Cell index.
    """
    editor.setGeometry(option.rect)                                               # type: ignore[attr-defined]
    return



""" widget that shows the table of the items """
import itertools
import logging
import re
from enum import Enum
from pathlib import Path
from typing import Any
import pandas as pd
from PySide6.QtCore import QModelIndex, QSortFilterProxyModel, Qt, Slot  # pylint: disable=no-name-in-module
from PySide6.QtGui import (QRegularExpressionValidator, QStandardItem,  # pylint: disable=no-name-in-module
                           QStandardItemModel)
from PySide6.QtWidgets import (QApplication, QComboBox, QFileDialog, QHeaderView,  # pylint: disable=no-name-in-module
                               QLineEdit, QMenu, QMessageBox, QPushButton, QTableView, QVBoxLayout, QWidget)
from ..guiCommunicate import Communicate
from ..guiStyle import Action, IconButton, Label, TextButton, space, widgetAndLayout, widgetAndLayoutGrid
from ..miscTools import callAddOn
from .gallery import ImageGallery
from .tableHeader import TableHeader


#Scan button to more button
class Table(QWidget):
  """ widget that shows the table of the items """
  def __init__(self, comm:Communicate):
    """
    Initialization

    Args:
      comm (Communicate): communication channel
    """
    super().__init__()
    self.comm = comm
    comm.changeTable.connect(self.change)
    comm.stopSequentialEdit.connect(self.stopSequentialEditFunction)
    self.stopSequentialEdit = False
    self.data         :pd.DataFrame = pd.DataFrame()
    self.models       :list[QStandardItemModel | QSortFilterProxyModel] = []
    self.filterSelect :list[QComboBox]   = []
    self.filterText   :list[QLineEdit]   = []
    self.filterInverse:list[QPushButton] = []
    self.filterDelete :list[QPushButton] = []
    self.docType = ''
    self.projID = ''
    self.filterHeader:list[str] = []
    self.showAll= True
    self.lastClickedRow = -1
    self.flagGallery = False

    ### GUI elements
    mainL = QVBoxLayout()
    mainL.setSpacing(0)
    mainL.setContentsMargins(space['s'], space['s'], space['s'], space['s'])
    # header
    self.headerW, headerL = widgetAndLayout('H', mainL, 'm')
    self.headerW.hide()
    self.headline = Label('','h1', headerL)
    self.showState= Label('', 'h3', headerL)
    self.subDocTypeL= Label('      subType:', 'h3', headerL)
    self.subDocTypeL.hide()
    self.subDocType=QComboBox(self)
    self.subDocType.setMinimumWidth(200)
    self.subDocType.setStyleSheet(self.comm.palette.get('secondaryText','color'))
    self.subDocType.currentTextChanged.connect(lambda dt: self.change(dt, self.projID))
    self.subDocType.hide()
    headerL.addWidget(self.subDocType)
    headerL.addStretch(1)
    self.addBtn = TextButton('Add',  self, [Command.ADD_ITEM],   headerL)

    self.selectionBtn = TextButton('Selection', self, [], headerL)
    selectionMenu = QMenu(self)
    Action('Toggle selection',self, [Command.TOGGLE_SELECTION], selectionMenu)
    selectionMenu.addSeparator()
    Action('Group Edit',      self, [Command.GROUP_EDIT],       selectionMenu)
    Action('Sequential edit', self, [Command.SEQUENTIAL_EDIT],  selectionMenu)
    Action('Rerun extractors',self, [Command.RERUN_EXTRACTORS], selectionMenu)
    Action('Delete',          self, [Command.DELETE],           selectionMenu)
    self.selectionBtn.setMenu(selectionMenu)

    self.visibilityBtn = TextButton('Visibility', self, [], headerL)
    visibilityMenu = QMenu(self)
    Action(                    'Add Filter',                        self, [Command.ADD_FILTER],     visibilityMenu)
    self.showHidden    = Action('Show/hide hidden ___',             self, [Command.SHOW_ALL],       visibilityMenu)
    self.toggleHidden  = Action('Invert hidden status of selected', self, [Command.TOGGLE_HIDE],    visibilityMenu)
    self.toggleGallery = Action('Toggle gallery view',              self, [Command.TOGGLE_GALLERY], visibilityMenu)
    self.toggleGallery.setVisible(False)
    self.visibilityBtn.setMenu(visibilityMenu)

    more = TextButton('More', self, [], headerL)
    self.moreMenu = QMenu(self)
    Action('Export to csv',            self, [Command.EXPORT],   self.moreMenu)
    self.moreMenu.addSeparator()
    projectGroup = self.comm.backend.configuration['projectGroups'][self.comm.backend.configurationProjectGroup]
    if projectAddOns := projectGroup.get('addOns',{}).get('table',''):
      for label, description in projectAddOns.items():
        Action(description, self, [Command.ADD_ON, label], self.moreMenu)
      self.moreMenu.addSeparator()
    self.actionChangeColums = Action('Change columns',  self, [Command.CHANGE_COLUMNS], self.moreMenu)  #add this action at end
    more.setMenu(self.moreMenu)

    # filter
    _, self.filterL = widgetAndLayoutGrid(mainL, top='s', bottom='s')
    # table
    self.table = QTableView(self)
    self.table.setStyleSheet('QTableView::indicator {width: 24px; height: 24px;}')
    self.table.verticalHeader().hide()
    self.table.clicked.connect(self.cellClicked)
    self.table.doubleClicked.connect(self.cell2Clicked)
    self.table.setSortingEnabled(True)
    self.table.setAlternatingRowColors(True)
    header = self.table.horizontalHeader()
    header.setStyleSheet('QHeaderView::section {padding: 0px 5px; margin: 0px;}')
    header.setSectionsMovable(True)
    header.setSortIndicatorShown(True)
    header.setMaximumSectionSize(self.comm.backend.configuration['GUI']['maxTableColumnWidth'])
    header.setStretchLastSection(True)
    mainL.addWidget(self.table)
    self.gallery = ImageGallery(self.comm)
    self.gallery.setVisible(False)
    mainL.addWidget(self.gallery)
    self.setLayout(mainL)


  @Slot(str, str)                                       # type: ignore[arg-type]
  def change(self, docType:str, projID:str) -> None:
    """
    What happens when the table changes its raw information

    Args:
      docType (str): document type; leave empty for redraw
      projID (str): id of project
    """
    if docType!=self.docType or projID!=self.projID:
      logging.debug('table:changeTable |%s|%s|',docType, projID)
    self.models = []
    self.filterSelect = []
    self.filterText   = []
    self.filterInverse= []
    self.filterDelete = []
    #if not docType:  #only remove old filters, when docType changes
    #   make sure internal updates are accounted for: i.e. comment
    for i in reversed(range(self.filterL.count())):
      item_1 = self.filterL.itemAt(i)
      if item_1 is not None:
        item_1.widget().setParent(None)
    allDocTypes:list[str] = []
    if (docType!=self.docType or projID!=self.projID) and '/' not in docType:
      # get list of all subDocTypes
      allDocTypes = self.comm.backend.db.dataHierarchy('', '')
      allDocTypes = [i for i in allDocTypes if i.startswith(docType)]
      if len(allDocTypes)==1:
        self.subDocTypeL.hide()
        self.subDocType.hide()
      elif len(allDocTypes)>1 and docType:
        self.subDocTypeL.show()
        self.subDocType.show()
        alreadyInside = {self.subDocType.itemText(i) for i in range(self.subDocType.count())}
        if alreadyInside != set(allDocTypes):
          self.subDocType.clear()
          self.subDocType.addItems(allDocTypes)
    if docType!='':
      self.docType = docType
      self.projID  = projID
    if 'measurement' in self.docType:
      self.toggleGallery.setVisible(True)
    else:
      self.toggleGallery.setVisible(False)
      self.flagGallery = False
    # start tables: collect data
    if self.docType=='_tags_':
      self.addBtn.hide()
      if self.showAll:
        self.data = self.comm.backend.db.getView('viewIdentify/viewTagsAll')
      else:
        self.data = self.comm.backend.db.getView('viewIdentify/viewTags')
      self.filterHeader = ['tag','name','type']
      self.headline.setText('TAGS')
      self.actionChangeColums.setVisible(False)
    else:
      self.addBtn.show()
      if self.docType.startswith('x0'):
        self.selectionBtn.hide()
        self.toggleHidden.setVisible(False)
      else:
        self.selectionBtn.show()
        self.toggleHidden.setVisible(True)
      path = (f'viewDocType/{self.docType}All' if self.showAll else f'viewDocType/{self.docType}')
      if self.projID=='':
        self.data = self.comm.backend.db.getView(path)
      else:
        self.data = self.comm.backend.db.getView(path, startKey=self.projID)
      self.showState.setText('(show all rows)' if self.showAll else '(hide hidden rows)')
      docLabel = 'Unidentified'
      if self.docType=='-':
        self.actionChangeColums.setVisible(False)
      else:
        self.actionChangeColums.setVisible(True)
        docLabel = self.comm.backend.db.dataHierarchy(self.docType,'title')[0]
      if self.projID:
        self.headline.setText(docLabel)
        self.showHidden.setText(f'Show/hide hidden {docLabel.lower()}')
      else:
        self.comm.changeSidebar.emit('')  #close the project in sidebar
        self.headline.setText(f'All {docLabel.lower()}')
        self.showHidden.setText(f'Show/hide all hidden {docLabel.lower()}')
      self.filterHeader = list(self.data.columns)[:-1]
    self.headerW.show()
    nRows, nCols = self.data.shape
    model = QStandardItemModel(nRows, nCols-1)
    model.setHorizontalHeaderLabels(self.filterHeader)
    for i, j in itertools.product(range(nRows), range(nCols-1)):
      value = self.data.iloc[i,j]
      if self.docType=='_tags_':  #tags list
        if j==0:
          if value=='_curated':          # curated
            item = QStandardItem('_curated_')
          elif re.match(r'_\d', value):  # star
            item = QStandardItem('\u2605'*int(value[1]))
          else:
            item = QStandardItem(value)
        else:
          item = QStandardItem(value)
      elif value in ('None','','nan'):  #None, False
        item = QStandardItem('-') # if you want to add nice glyphs, see also below \u00D7')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      elif value=='True': #True
        item = QStandardItem('Y') # \u2713')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      elif isinstance(value, str) and re.match(r'^[a-z\-]-[a-z0-9]{32}$',value):      #Link
        item = QStandardItem('oo') # \u260D')
        # item.setFont(QFont("Helvetica [Cronyx]", 16))
      else:
        if self.filterHeader[j]=='tags':
          tags = [i.strip() for i in value.split(',')]
          if '_curated' in tags:
            tags[tags.index('_curated')] = '_curated_'
          elementStar = list(filter(re.compile(r'^_\d$').match, tags))
          if elementStar:
            tags.remove(elementStar[0])
            tags = ['\u2605'*int(elementStar[0][1])]+tags
          text = ' '.join(tags)
        else:
          text = value
        item = QStandardItem(text)
      if j == 0:
        doc = self.comm.backend.db.getDoc(self.data['id'][i])
        if [b for b in doc['branch'] if False in b['show']]:
          item.setText(f'{item.text()}  \U0001F441')
        item.setAccessibleText(doc['id'])
        if self.docType != 'x0':
          item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
          item.setCheckState(Qt.CheckState.Unchecked)
      else:
        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable)
      model.setItem(i, j, item)
    self.models = [model]
    if self.flagGallery:
      self.gallery.updateGrid(model)
      self.gallery.setVisible(True)
      self.table.setVisible(False)
    else:
      self.table.setModel(self.models[-1])
      self.table.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
      self.table.horizontalHeader().setStretchLastSection(True)
      self.table.sortByColumn(0, Qt.SortOrder.AscendingOrder)
      self.table.show()
      self.gallery.setVisible(False)
      self.table.setVisible(True)
    return


  def execute(self, command:list[Any]) -> None:
    """
    Event if user clicks button in the center

    Args:
      command (list): list of commands
    """
    if command[0] is Command.ADD_ITEM:
      self.comm.formDoc.emit({'type':[self.docType], '_projectID':self.projID})
      self.comm.changeTable.emit(self.docType, self.projID)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')
    elif command[0] is Command.GROUP_EDIT:
      intersection = None
      docIDs = []
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          docIDs.append(docID)
          thisKeys = set(self.comm.backend.db.getDoc(docID))
          if intersection is None:
            intersection = thisKeys
          else:
            intersection = intersection.intersection(thisKeys)
      #remove keys that should not be group edited and build dict
      if intersection is not None:
        intersection = intersection.difference({'branch', 'user', 'client', 'metaVendor', 'shasum', \
            'id', 'metaUser', 'rev', 'name', 'dateCreated', 'dateModified', 'image', 'links', 'gui'})
        intersectionDict:dict[str,Any] = {i:'' for i in intersection}  | \
                                         {k:v  for k,v in self.comm.backend.db.getDoc(docID).items()
                                          if isinstance(v,dict) and k not in ('metaUser','metaVendor')}
        intersectionDict['tags'] = []
        intersectionDict['type'] = [self.docType]
        intersectionDict['_ids'] = docIDs
        self.comm.formDoc.emit(intersectionDict)
        self.comm.changeDetails.emit('redraw')
        self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.SEQUENTIAL_EDIT:
      self.stopSequentialEdit = False
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.formDoc.emit(self.comm.backend.db.getDoc(docID))
        if self.stopSequentialEdit:
          break
      self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.DELETE:
      ret = None
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          if ret is None:
            ret = QMessageBox.critical(self, 'Warning', 'Are you sure you want to delete the data?',
                QMessageBox.StandardButton.Yes, QMessageBox.StandardButton.No)
          if ret==QMessageBox.StandardButton.Yes:
            doc = self.comm.backend.db.getDoc(docID)
            for branch in doc['branch']:
              if branch['path'] is not None:
                oldPath = self.comm.backend.basePath/branch['path']
                if oldPath.exists():
                  newPath = oldPath.parent / f'trash_{oldPath.name}'
                  oldPath.rename(newPath)
            self.comm.backend.db.remove(docID)
      self.comm.changeTable.emit(self.docType, self.projID)
    elif command[0] is Command.CHANGE_COLUMNS:
      dialog = TableHeader(self.comm, self.docType)
      dialog.exec()
    elif command[0] is Command.EXPORT:
      fileName = QFileDialog.getSaveFileName(self,'Export to ..',str(Path.home()),'*.csv')[0]
      if not fileName.endswith('.csv'):
        fileName += '.csv'
      with open(fileName,'w', encoding='utf-8') as fOut:
        header = [f'"{i}"' for i in self.filterHeader]
        fOut.write(','.join(header)+'\n')
        for row in range(self.models[-1].rowCount()):
          rowContent = []
          for col in range(self.models[-1].columnCount()):
            value = self.models[-1].index( row, col, QModelIndex()).data(Qt.ItemDataRole.DisplayRole)
            rowContent.append(f'"{value}"')
          fOut.write(','.join(rowContent)+'\n')
    elif command[0] is Command.ADD_ON:
      data   = []
      for row in range(self.models[-1].rowCount()):
        _, docID = self.itemFromRow(row)
        path = self.comm.backend.db.getDoc(docID)['branch'][0]['path']
        dataRow = [docID, '' if path is None else str(self.comm.backend.basePath/path)]
        for col in range(self.models[-1].columnCount()):
          value = self.models[-1].index(row, col).data(Qt.ItemDataRole.DisplayRole)
          dataRow.append(value)
        data.append(dataRow)
      df = pd.DataFrame(data, columns=['docID','path']+self.filterHeader)
      callAddOn(command[1], self.comm.backend, df, self)
    elif command[0] is Command.TOGGLE_HIDE:
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          self.comm.backend.db.hideShow(docID)
      if self.docType=='x0':
        self.comm.changeSidebar.emit('redraw')
      self.change('','')  # redraw table
    elif command[0] is Command.TOGGLE_SELECTION:
      for row in range(self.models[-1].rowCount()):
        item,_ = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          item.setCheckState(Qt.CheckState.Unchecked)
        else:
          item.setCheckState(Qt.CheckState.Checked)
    elif command[0] is Command.SHOW_ALL:
      self.showAll = not self.showAll
      self.change('','')  # redraw table
    elif command[0] is Command.RERUN_EXTRACTORS:
      redraw = False
      for row in range(self.models[-1].rowCount()):
        item, docID = self.itemFromRow(row)
        if item.checkState() == Qt.CheckState.Checked:
          doc = self.comm.backend.db.getDoc(docID)
          if doc['branch'][0]['path'] is None:
            continue
          redraw = True
          oldDocType = doc['type']
          doc['type'] = ['']
          if doc['branch'][0]['path'].startswith('http'):
            path = Path(doc['branch'][0]['path'])
          else:
            path = self.comm.backend.basePath/doc['branch'][0]['path']
          self.comm.backend.useExtractors(path, doc.get('shasum',''), doc)
          if doc['type'][0] == oldDocType[0]:
            del doc['branch']  #don't update
            self.comm.backend.db.updateDoc(doc, docID)
          else:
            self.comm.backend.db.remove( docID )
            del doc['id']
            doc['name'] = doc['branch'][0]['path']
            self.comm.backend.addData('/'.join(doc['type']), doc, doc['branch'][0]['stack'])
      if redraw:
        self.change('','')  # redraw table
        self.comm.changeDetails.emit('redraw')
    elif command[0] is Command.TOGGLE_GALLERY:
      self.flagGallery = not self.flagGallery
      self.change('','')    # redraw table/gallery
    elif command[0] is Command.ADD_FILTER:
      # gui
      _, rowL = widgetAndLayout('H', self.filterL, 'm', 'xl', '0', 'xl')
      self.filterSelect.append(QComboBox())
      self.filterSelect[-1].setStyleSheet(self.comm.palette.get('secondaryText', 'color'))
      self.filterSelect[-1].addItems(self.filterHeader+['rating'] if 'tag' in self.filterHeader else
                                     self.filterHeader)
      self.filterSelect[-1].currentIndexChanged.connect(self.filterChoice)
      self.filterSelect[-1].setMinimumWidth(max(len(i) for i in self.filterHeader)*14)
      rowL.addWidget(self.filterSelect[-1])
      self.filterText.append(QLineEdit(''))
      self.filterText[-1].setStyleSheet(self.comm.palette.get('secondaryText', 'color'))
      self.filterText[-1].setValidator(QRegularExpressionValidator(r'[a-zA-Z0-9_\.]+'))
      rowL.addWidget(self.filterText[-1])
      btnInverse = IconButton('ph.selection-inverse-fill', self, [Command.SET_FILTER,    len(self.models), 'invert'], rowL, checkable=True)
      self.filterInverse.append(btnInverse)
      btnDelete  = IconButton('fa5s.minus-square',         self, [Command.DELETE_FILTER, len(self.models)],           rowL)
      self.filterDelete.append(btnDelete)

      # data
      filterModel = QSortFilterProxyModel()
      self.filterText[-1].textChanged.connect(self.filterTextChanged)
      filterModel.setSourceModel(self.models[-1])
      filterModel.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
      filterModel.setFilterKeyColumn(0)
      self.models.append(filterModel)
      self.table.setModel(self.models[-1])
    elif command[0] is Command.DELETE_FILTER: # Remove filter from list of filters
      row = command[1]
      # change the information in the minus-button command
      for i in range(row, self.filterL.count()):        #e.g. model 1 is in row=0, so start in 1 for renumbering
        item_1 = self.filterL.itemAt(i)
        if item_1 is not None:
          layout_2 = item_1.widget().layout()
          if layout_2 is not None:
            item_2   = layout_2.itemAt(2)
            if item_2 is not None:
              item_2.widget().command[1] -= 1   # type: ignore[attr-defined]
      # remove row in GUI
      item_1 = self.filterL.itemAt(row-1)
      if item_1 is not None:
        item_1.widget().setParent(None) # e.g. model 1 is in row=0 for deletion
      # delete one row in models and adopt the sources
      del self.models[row]
      for i in range(1, len(self.models)):
        self.models[i].setSourceModel(self.models[i-1])                             # type: ignore[union-attr]
      self.table.setModel(self.models[-1])
    elif command[0] is Command.SET_FILTER:
      self.filterTextChanged('', command[1])
    else:
      print('**ERROR table menu unknown:',command)
    return


  def filterTextChanged(self, _:Any, row:int=-1) -> None:
    """ text in line-edit in the filter is changed: update regex

    Args:
      row (int): row number
    """
    if row<0:
      for idx,lineEdit in enumerate(self.filterText):
        if lineEdit==self.sender():
          row = idx
    else:
      row -= 1
    regexStr = self.filterText[row].text()
    if '*' in regexStr:
      regexStr = regexStr.replace('*','\u2605')
      if self.filterInverse[row].isChecked():
        regexStr = f'^((?!{regexStr}).)*$'
      else:
        regexStr = f'^{regexStr}$'
    elif self.filterInverse[row].isChecked():
      regexStr = f'^((?!{regexStr}).)*$'
    self.models[row+1].setFilterRegularExpression(regexStr)                         # type: ignore[union-attr]
    return


  def cellClicked(self, index: QModelIndex) -> None:
    """
    What happens when user clicks cell in table of tags, projects, samples, ...
    -> show details

    Args:
      index (QModelIndex): cell clicked
    """
    row = index.row()
    _, docID = self.itemFromRow(row)

    # Check if shift is held and lastClickedRow is set
    modifiers = QApplication.keyboardModifiers()
    if modifiers == Qt.ShiftModifier and self.lastClickedRow > -1:   # type: ignore[attr-defined]
      start = min(self.lastClickedRow, row)
      end = max(self.lastClickedRow, row)
      target_state = Qt.CheckState.Checked if self.itemFromRow(row)[0].checkState() == Qt.CheckState.Checked \
                      else Qt.CheckState.Unchecked
      for r in range(start, end + 1):
        item, _ = self.itemFromRow(r)
        item.setCheckState(target_state)
    else: # No need to toggle only the clicked row, just record it
      self.lastClickedRow = row

    if docID[0] == 'x':  # only show items for non-folders
      doc = self.comm.backend.db.getDoc(docID)
      if doc['type'][0] == 'x0':
        self.comm.changeProject.emit(docID, '')
        self.comm.changeSidebar.emit(docID)
      else:
        projID = doc['branch'][0]['stack'][0]
        self.comm.changeProject.emit(projID, docID)
        self.comm.changeSidebar.emit(projID)
    else:
      self.comm.changeDetails.emit(docID)
    return


  def cell2Clicked(self, index: QModelIndex) -> None:
    """
    What happens when user double clicks cell in table of projects

    Args:
      index (QModelIndex): cell clicked
    """
    row = index.row()
    _, docID = self.itemFromRow(row)
    if self.docType=='x0':
      self.comm.changeProject.emit(docID, '')
      self.comm.changeSidebar.emit(docID)
    else:
      doc = self.comm.backend.db.getDoc(docID)
      self.comm.formDoc.emit(doc)
      self.comm.changeTable.emit('','')
      self.comm.changeDetails.emit('redraw')
    return


  @Slot()
  def stopSequentialEditFunction(self) -> None:
    """ Stop the sequential edit of a number of items """
    self.stopSequentialEdit=True
    return


  def itemFromRow(self, row:int) -> tuple[QStandardItem, str]:
    """
    get item from row by iterating through the proxyModels

    Args:
      row (int): row number

    Returns:
      QItem, str: the item and docID
    """
    index = self.models[-1].index(row,0)
    for idxModel in range(len(self.models)-1,0,-1):
      index = self.models[idxModel].mapToSource(index)                              # type: ignore[union-attr]
    item = self.models[0].itemFromIndex(index)                                      # type: ignore[union-attr]
    return item, item.accessibleText()


  def filterChoice(self, item:int) -> None:
    """
    Change the column which is used for filtering

    Args:
       item (int): column number to filter by
    """
    rating = False
    if item == self.models[-1].columnCount():  # ratings
      item = self.filterHeader.index('tag')
      rating = True
    for idx,combobox in enumerate(self.filterSelect):
      if combobox==self.sender():
        self.models[idx+1].setFilterKeyColumn(item)                                 # type: ignore[union-attr]
        self.filterText[idx].setText('')
        if rating:
          self.filterText[idx].setValidator(QRegularExpressionValidator(r'\*+'))
        else:
          self.filterText[idx].setValidator(QRegularExpressionValidator(r'[a-zA-Z0-9_]+'))
    return


class Command(Enum):
  """ Commands used in this file """
  ADD_FILTER       = 1
  ADD_ITEM         = 2
  TOGGLE_SELECTION = 3
  GROUP_EDIT       = 4
  SEQUENTIAL_EDIT  = 5
  TOGGLE_HIDE      = 6
  RERUN_EXTRACTORS = 7
  DELETE           = 8
  SHOW_ALL         = 9
  EXPORT           = 10
  CHANGE_COLUMNS   = 11
  DELETE_FILTER    = 12
  SET_FILTER       = 13
  TOGGLE_GALLERY   = 14
  ADD_ON           = 15



""" Displays a scrollable grid of images (PNG, JPG, SVG) """
from PySide6.QtCore import QByteArray, Qt, Signal
from PySide6.QtGui import QImage, QMouseEvent, QPixmap, QStandardItemModel
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import QGridLayout, QLabel, QPushButton, QScrollArea, QVBoxLayout, QWidget
from ..guiCommunicate import Communicate

IMG_SIZE = 300

class ClickableImage(QLabel):
  """
  A QLabel subclass that emits signals when clicked or double-clicked
  """
  clicked = Signal(str)
  doubleClicked = Signal(str)

  def __init__(self, docID: str):
    """
    Initializes the ClickableImage

    Args:
      docID: The unique identifier for the image
    """
    super().__init__()
    self.docID = docID


  def mousePressEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse press events. Emits 'clicked' signal on left-button press

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.clicked.emit(self.docID)
    return


  def mouseDoubleClickEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse double-click events. Emits 'doubleClicked' signal on left-button double-click

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.doubleClicked.emit(self.docID)
    return




class ClickableSvgButton(QPushButton):
  """
  A QPushButton subclass specifically for displaying SVG images that
  emits a signal when double-clicked
  #TODO merge with ClickableImage and Image (guiStyle)
  """
  clicked = Signal(str)
  doubleClicked = Signal(str)

  def __init__(self, docID:str):
    """
    Initializes the ClickableSvgButton

    Args:
      docID: The unique identifier for the SVG image
    """
    super().__init__()
    self.docID = docID
    self.setFixedSize(200, 200)
    self.setStyleSheet('border: none;')


  def mousePressEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse press events. Emits 'clicked' signal on left-button press

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.clicked.emit(self.docID)
    return


  def mouseDoubleClickEvent(self, event:QMouseEvent) -> None:
    """
    Handles mouse double-click events. Emits 'doubleClicked' signal on left-button double-click

    Args:
      event: The QMouseEvent
    """
    if event.button() == Qt.LeftButton:                                           # type: ignore[attr-defined]
      self.doubleClicked.emit(self.docID)
    return




class ImageGallery(QWidget):
  """
  A widget that displays a scrollable grid of images (PNG, JPG, SVG)
  Images can be clicked or double-clicked to trigger actions
  """
  def __init__(self, comm:Communicate):
    """
    Initializes the ImageGallery

    Args:
      comm: The communication object used to interact with the backend and other UI components
    """
    super().__init__()
    self.comm = comm
    layout = QVBoxLayout(self)
    scrollArea = QScrollArea(self)
    scrollArea.setWidgetResizable(True)
    scrollContent = QWidget()
    self.gridL = QGridLayout(scrollContent)
    scrollArea.setWidget(scrollContent)
    layout.addWidget(scrollArea)

  def updateGrid(self, model:QStandardItemModel) -> None:
    """
    Populates or updates the image grid based on the provided model

    Clears the existing grid and then iterates through the model to fetch
    and display images. It handles both SVG and raster image formats (like PNG, JPG)

    Args:
      model: The data model containing information about the images to display.
             The first column of each row is expected to have an 'accessibleText'
    """
    # Clear existing widgets from the grid
    while self.gridL.count():
      child = self.gridL.takeAt(0)
      if child.widget():
        child.widget().deleteLater()

    row, col = 0, 0
    for idx in range(model.rowCount()):
      docID = model.itemFromIndex(model.index(idx,0)).accessibleText()
      image_data = self.comm.backend.db.getDoc(docID)
      if not image_data or 'image' not in image_data:
        # Handle cases where image data might be missing or malformed
        print(f"Warning: Image data not found or malformed for docID: {docID}")
        continue
      image = image_data['image']

      if image.startswith('<?xml version="1.0"'):
        # Wrap QSvgWidget in a Clickable subclass of QPushButton
        button = ClickableSvgButton(docID)
        button.setFixedSize(IMG_SIZE, IMG_SIZE)
        # button.setStyleSheet("border: none;") # Already set in ClickableSvgButton.__init__
        svgWidget = QSvgWidget()
        svgWidget.renderer().load(bytearray(image, encoding='utf-8'))
        # Calculate aspect ratio for SVG
        svg_width = svgWidget.renderer().defaultSize().width()
        svg_height = svgWidget.renderer().defaultSize().height()

        if svg_width == 0 or svg_height == 0: # Avoid division by zero
          new_width, new_height = IMG_SIZE -4, IMG_SIZE -4
        elif svg_height > svg_width:
          new_width = int(float(svg_width) / float(svg_height) * (IMG_SIZE - 4))
          new_height = IMG_SIZE - 4
        else:
          new_width = IMG_SIZE - 4
          new_height = int(float(svg_height) / float(svg_width) * (IMG_SIZE - 4))
        svgWidget.setMaximumSize(new_width, new_height)

        layout = QVBoxLayout(button)
        layout.addWidget(svgWidget)
        layout.setAlignment(Qt.AlignCenter)                                       # type: ignore[attr-defined]
        button.clicked.connect(lambda checked=False, docID=docID: self.imageClicked(docID))
        button.doubleClicked.connect(self.image2Clicked)
        self.gridL.addWidget(button, row, col)
      elif 'base64,' in image:  # Basic check for base64 data URI
        try:
          header, base64_data = image.split(',', 1)
          # Extract image type from header (e.g., "data:image/png;base64")
          img_type_part = header.split(';')[0].split('/')[-1]
          imageType = img_type_part.upper()

          byteArr = QByteArray.fromBase64(bytearray(base64_data, encoding='utf-8'))
          imageW = QImage()
          # Ensure format string is clean (e.g. PNG, JPEG)
          fmt = imageType.replace(';', '')
          if imageW.loadFromData(byteArr, format=fmt):
            pixmap = QPixmap.fromImage(imageW).scaled(IMG_SIZE,IMG_SIZE,Qt.KeepAspectRatio,Qt.SmoothTransformation) # type: ignore[attr-defined]
            label = ClickableImage(docID)
            label.setPixmap(pixmap)
            label.setAlignment(Qt.AlignCenter)                            # type: ignore[attr-defined]
            label.clicked.connect(self.imageClicked)
            label.doubleClicked.connect(self.image2Clicked)
            self.gridL.addWidget(label, row, col)
          else:
            #TODO next lines should be in details
            print(f"Warning: Could not load image data for docID: {docID} with format {fmt}")
        except ValueError:
          print(f"Warning: Malformed base64 image data URI for docID: {docID}")
        except Exception as e:
          print(f"Error processing image for docID {docID}: {e}")
      else:
        print(f"Warning: Image for docID {docID} is not in expected base64 format.")
      # end of loop
      col += 1
      if col >= 4: # Assuming 4 columns
        col = 0
        row += 1
    return


  def imageClicked(self, docID:str) -> None:
    """
    Handles single-click events on an image

    Args:
      docID: The unique identifier of the clicked image/document
    """
    self.comm.changeDetails.emit(docID)
    return


  def image2Clicked(self, docID:str) -> None:
    """
    Handles double-click events on an image

    Args:
      docID: The unique identifier of the double-clicked image/document
    """
    doc = self.comm.backend.db.getDoc(docID)
    self.comm.formDoc.emit(doc)
    return



#!/usr/bin/python3
"""Commandline utility to admin local installation and convert from Pasta-ELN version 2"""
import json
import os
import platform
import re
import shutil
import sys
import traceback
from pathlib import Path
from sqlite3 import IntegrityError
from typing import Any, Callable, Union
import requests
from requests.structures import CaseInsensitiveDict
from pasta_eln.backend import Backend
from pasta_eln.elabFTWsync import Pasta2Elab
from pasta_eln.fixedStringsJson import CONF_FILE_NAME, defaultDocTypes, defaultSchema
from pasta_eln.textTools.stringChanges import outputString


class Tools:
  """Commandline utility to admin local installation and convert from Pasta-ELN version 2"""
  def __init__(self) -> None:
    self.backend:Backend|None = None
    self.projectGroup = ''


  def __choice__(self, command:str) -> str:
    """  depending on choice, execute command
    Args:
      command (str): command
    """
    helpString = 'Command line program to troubleshoot and possibly repair Pasta-ELN database\n'\
                'Commands - general:\n'\
                '  [h]elp: long help\n'\
                '  [q]uit\n'
    helpString += '  [p]rint a document\n'
    if command == 'p':
      self.printOrDelete()
    helpString += '  [d]elete a document\n'
    if command == 'd':
      self.printOrDelete(projectGroup='', docID='', output=False)
    helpString += '  [s]can all projects\n'
    if command == 's':
      self.scanAllProjects()

    helpString += 'Commands - update Version 2 -> Version 3:\n'
    helpString += '  [u1] convert couchDB to SQLite\n'
    if command == 'u1':
      self.couchDB2SQLite()
    helpString += '  [u2] update disk structure from V2->v3\n'
    if command == 'u2':
      self.translateV2_V3()

    helpString += 'Commands - update:\n'
    helpString += '  [u3] convert from one procedure to 3 workflows\n'
    if command == 'u3':
      self.update3Workflow()

    helpString += 'Commands - database integrity:\n'
    helpString += '  [v]erify\n'
    if command == 'v':
      self.verifyPasta()
    helpString += '  [r]epair\n'
    if command == 'r':
      self.verifyPasta('', self.__userQuestion__)
    helpString += '  [rA]epair: answer always "yes": WARNING can change too much\n'
    if command == 'rA':
      self.verifyPasta('', lambda t: self.__userQuestion__(t, True))
    helpString += '  [rS]repair schema: reset the doc-type schema to its original state\n'
    if command == 'rS':
      self.repairSchema()
    helpString += '  [rp1] repair properties: add missing "."\n'
    if command == 'rp1':
      self.repairPropertiesDot()
    helpString += '  [cp]-create a lost and found project: helpful for some repair operations\n'
    if command == 'cp':
      self.createLostAndFound()

    helpString += 'Commands - clone from / to server:\n'
    helpString += '  [ss]ync SEND ALL\n'
    helpString += '  [sg]ync GET ALL\n'
    if command in {'ss','sg'}:
      self.sync('', command)
    helpString += '  [pL]urge local database: REMOVE EVERYTHING\n'
    if command == 'pL':
      self.purgeLocal()
    helpString += '  [pR]urge remote database: REMOVE EVERYTHING\n'
    if command == 'pR':
      self.purgeRemote()

    return helpString if command=='h' else 'Done'


  def __setBackend__(self, projectGroup:str='') -> None:
    """ Internal function: return backend based on project group
    Args:
      projectGroup (str): "name" of project group
    """
    with open(Path.home()/CONF_FILE_NAME, encoding='utf-8') as fIn:
      config = json.load(fIn)
      print('Project groups:','  '.join(f'{idx+1}-{i}' for idx,i in enumerate(config['projectGroups'].keys())))
    while not self.projectGroup:
      projectGroup = projectGroup or input('  Enter project group [number or name or entire text]: ').strip()
      if re.match(r'^\d+-\w+$', projectGroup):
        self.projectGroup = projectGroup.split('-')[1]
      elif re.match(r'^\d+$', projectGroup):
        self.projectGroup = list(config['projectGroups'].keys())[int(projectGroup)-1]
      elif projectGroup in config['projectGroups'].keys():
        self.projectGroup = projectGroup
      else:
        projectGroup = ''
    self.backend = Backend(self.projectGroup)
    return


  def __progressBar__(self, iteration:int, total:int=100, prefix:str='', suffix:str='', decimals:int=1,
                      length:int=100, fill:str='█', printEnd:str='') -> None:
    """
    Call in a loop to create terminal progress bar
    - https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters

    Args:
        iteration (int): current iteration
        total     (int): total iterations
        prefix    (str): prefix string
        suffix    (str): suffix string
        decimals  (int): positive number of decimals in percent complete
        length    (int): character length of bar
        fill      (int): bar fill character
        printEnd  (int): end character (e.g. "\r", "\r\n")
    """
    percent = ('{0:.' + str(decimals) + 'f}').format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    barString = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{barString}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total:
      print()
    return


  def __updateProgressBar__(self, dType:str, data:str|int) -> None:
    """ update dialog
    - "text" and "append" will update the text

    Args:
      dType (str): what to update and how "text", "append", "count", "incr"
      data (str): str- or int-value to update with
    """
    if dType in {'text','append'} and isinstance(data, str):
      print(data)
    if dType=='count' and isinstance(data, int):
      self.__progressBar__(data)
    return


  def __userQuestion__(self, errorMessage:str, allYes:bool=False) -> bool:
    """ Ask user if repair should be done

    Args:
      errorMessage (str): error message
      allYes (bool): answer all questions with yes

    Return:
      bool: repair should be done
    """
    print(errorMessage)
    return True if allYes else input('Repair [yN]: ')=='y'


  def run(self, todos:list[str]=[]) -> None:
    """  run the list of given commands, else run while-loop until user ends
    Args:
      todos (list): string-commands
    """
    if todos:
      self.__setBackend__(todos.pop(0))
    while True:
      command = todos.pop(0) if todos else input('\nImportant commands [h]elp; [q]uit; [v]erify; [r]epair: ')
      if command == 'q':
        break
      print(self.__choice__(command))
    return


  def printOrDelete(self, projectGroup:str='', docID:str='', output:bool=True) -> None:
    """ Print or delete a document
    Args:
      projectGroup (str): from which project group to delete
      docID (str): which ID to delete
      output (bool): print=True or delete=False
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    self.backend = Backend(projectGroup)
    if not docID:
      docID = input('Enter docID: ').strip()
    if output:
      print(json.dumps(self.backend.db.getDoc(docID), indent=2))
    else:
      self.backend.db.remove(docID)
    return


  def scanAllProjects(self, projectGroup:str='') -> None:
    """ Scan all projects to identify missing items

      Args:
        projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    projectGroups = self.backend.db.getView('viewDocType/x0')['id'].values
    for projID in projectGroups:
      self.backend.scanProject(lambda i: self.__updateProgressBar__('count',i), projID)
    return


  def couchDB2SQLite(self, userName:str='', password:str='', database:str='', path:str='') -> None:
    """
    Backup everything of the CouchDB installation (local couchdb instance)

    Args:
      userName (str): username
      password (str): password
      database (str): database
      path     (str): path to location of sqlite file
    """
    headers:CaseInsensitiveDict[str]= CaseInsensitiveDict()
    headers['Content-Type'] = 'application/json'
    # get arguments if not given
    location = '127.0.0.1'
    if not userName:
      userName = input('Enter local admin username: ').strip()
    if not password:
      password = input('Enter password: ').strip()
    if not path:
      path = input('Enter path to data: ').strip()
    # use information
    authUser = requests.auth.HTTPBasicAuth(userName, password)
    resp = requests.get(f'http://{location}:5984/_all_dbs', headers=headers, auth=authUser, timeout=10)
    if resp.status_code != 200:
      print('**ERROR response for _all_dbs wrong', resp.text)
      print('Username and password', userName, password)
      return
    databases = resp.json()
    print('Databases:',databases)
    if not database:
      database = input('Enter database: ').strip()
    # big loop over all documents
    if self.backend is None:
      return
    db = self.backend.db
    resp = requests.get(f'http://{location}:5984/{database}/_all_docs', headers=headers, auth=authUser, timeout=10)
    for item in resp.json()['rows']:
      docID = item['id']
      if docID in ('-dataHierarchy-','-ontology-') or docID.startswith('_design/'):
        continue
      # print(f'DocID: {docID}')
      doc   = requests.get(f'http://{location}:5984/{database}/{docID}', headers=headers, auth=authUser, timeout=10).json()
      doc, attachments = self.translateDoc(doc)
      db.saveDoc(doc)
      for att in attachments:
        docAttach = requests.get(f'http://{location}:5984/{database}/{docID}/{att}',
                                  headers=headers, auth=authUser, timeout=10).json()
        setAll = set(docAttach.keys())
        setImportant  = setAll.difference({'-date','date','-client','image','type','-type','client','-name','-branch','branch'})
        if not setImportant or ('-name' in docAttach and docAttach['-name']=='new folder'):
          continue
        date = docAttach['-date'] if '-date' in docAttach else docAttach['date'] if 'date' in docAttach else att
        if '-client' in docAttach:
          del docAttach['-client']
        db.cursor.execute('INSERT INTO changes VALUES (?,?,?)', [docID, date, json.dumps(docAttach)])
        db.connection.commit()
    return


  def translateDoc(self, doc:dict[str,Any], comment:str='') -> tuple[dict[str,Any],list[Any]]:
    """ translate to new style

    Args:
      doc (dict): input document
      comment (str): comment to add to the output, e.g. offending doc-id

    Returns:
      dict: output document
    """
    from .textTools.handleDictionaries import fillDocBeforeCreate
    defaultValues = {'gui':[True,True], 'user':''}
    try:
      doc['id'] = doc.pop('_id')
      for key in ('name','user','type','gui','tags','client','branch','date'):
        if key in doc:      #skip if key is already in correct format
          continue
        if key in defaultValues and f'-{key}' not in doc:
          doc[key] = defaultValues[key]
          continue
        doc[key] = doc.pop(f'-{key}')
      doc['dateCreated']  = doc['date']
      doc['dateModified'] = doc.pop('date')
      doc['branch'] = doc['branch'][0] | {'op':'c'} if doc['branch'] else \
                      {'stack':[], 'child':9999, 'path':None, 'show':[True], 'op':'c'}
      del doc['_rev']
      attachments = doc.pop('_attachments',[])
      doc = fillDocBeforeCreate(doc, doc['type'])
    except Exception:
      print('Input document has mistakes in: ',comment,'\n',json.dumps(doc, indent=2))
      raise
    return doc, attachments


  def translateV2_V3(self, path:str='') -> None:
    """ Translate old id files in the file-tree into the new style
    Args:
      path (str): path to file-tree
    """
    if not path:
      path = input('Enter path to data: ').strip()
    for aPath, _, files in os.walk(path):
      if aPath == path or 'StandardOperatingProcedures' in aPath:
        continue
      if '.id_pastaELN.json' not in files:
        print('**ERROR** id file does NOT exist:', aPath,'\n   ',' '.join(files))
      with open(Path(aPath)/'.id_pastaELN.json', encoding='utf-8') as fIn:
        doc = json.load(fIn)
      docNew, _ = self.translateDoc(doc, aPath)
      del docNew['branch']['op']
      docNew['branch'] = [docNew['branch']]
      with open(Path(aPath)/'.id_pastaELN.json','w', encoding='utf-8') as fOut:
        fOut.write(json.dumps(docNew))
    return


  def update3Workflow(self, projectGroup:str='') -> None:
    """ Update from a 1 procedure into a 3 workflow (plan, procedure, log) version
    - Starting at 3.1 this new style is the default

    Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    cursor = self.backend.db.cursor

    cursor.execute("DELETE FROM docTypes WHERE docType = 'procedure'")
    for row in defaultDocTypes:
      if 'workflow' in row[0]:
        cmd = f"INSERT INTO docTypes VALUES ({', '.join(['?']*len(row))});"
        try:
          cursor.executemany(cmd, [row])
        except IntegrityError:
          print(f'Already up to date: docTypes {row[0]}')

    cursor.execute("DELETE FROM docTypeSchema WHERE docType = 'procedure'")
    for rowS in defaultSchema:
      if 'workflow' in str(rowS[0]):
        cmd = f"INSERT INTO docTypeSchema VALUES ({', '.join(['?']*len(rowS))});"
        try:
          cursor.executemany(cmd, [rowS])
        except IntegrityError:
          print(f'Already up to date: docTypeSchema {rowS[0]}')

    cursor.execute('SELECT type from main WHERE type LIKE "procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f'UPDATE main SET type="workflow/{docType}" WHERE type = "{docType}"'
      self.backend.db.cursor.execute(cmd)

    cursor.execute('SELECT key from properties WHERE key LIKE ".procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f"UPDATE properties SET key='.workflow/{docType[1:]}' WHERE key = '{docType}'"
      self.backend.db.cursor.execute(cmd)

    cursor.execute('SELECT key from definitions WHERE key LIKE "procedure%"')
    docTypesOld = {i[0] for i in cursor.fetchall()}
    for docType in docTypesOld:
      cmd =f"UPDATE definitions SET key='workflow/{docType}' WHERE key = '{docType}'"
      self.backend.db.cursor.execute(cmd)

    self.backend.db.connection.commit()
    return


  def verifyPasta(self, projectGroup:str='', repair:Union[None,Callable[[str],bool]]=None) -> None:
    """ Do the default verification of PastaELN. Adopted to CLI
      Args:
      projectGroup (str): name of project group
      repair (function): repair
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    outputString('print','info', self.backend.checkDB(outputStyle='text', repair=repair, minimal=False))
    # print hierarchy of project with name
    # projName = '...'
    # df = self.backend.db.getView('viewDocType/x0')
    # self.backend.changeHierarchy(df[df['name']==projName]['id'].values[0])
    # print(self.backend.outputHierarchy(False,True))
    return


  def repairSchema(self, projectGroup:str='') -> None:
    """ Do the default verification of PastaELN. Adopted to CLI
      Args:
      projectGroup (str): name of project group
      repair (function): repair
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    self.backend.db.dataHierarchyInit(True)
    return


  def createLostAndFound(self, projectGroup:str='') -> None:
    """ Create a lost and found project that is the target of misc documents
      Args:
        projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    self.backend.addData('x0', {'name': 'Lost and Found', 'comment': 'Location of lost database items'})
    return


  def purgeLocal(self, projectGroup:str='') -> None:
    """ Purge entire local storage, remove everything

      Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    dirName = self.backend.basePath
    if dirName.as_posix() != '/home/steffen/FZJ/pasta_misc/testing': #TODO temporary safeguard
      print('DO NOT DELETE THIS')
      return
    self.backend.exit()
    try:
      shutil.rmtree(dirName)
      os.makedirs(dirName)
      if platform.system()=='Windows':
        print('Try-Except unnecessary')
    except Exception:
      pass
    self.__setBackend__(self.projectGroup)
    return


  def purgeRemote(self, projectGroup:str='') -> None:
    """ Purge entire remote storage, remove everything

      Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    Pasta2Elab(self.backend, self.projectGroup, True)
    return



  def sync(self, projectGroup:str='', command:str='') -> None:
    """ synchronize with elab server

      Args:
      projectGroup (str): name of project group
      command (str): 'ss' or 'sg' to send/get data
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    syncObj = Pasta2Elab(self.backend, self.projectGroup)
    if command:
      syncObj.sync('sA' if command=='ss' else 'gA', progressCallback=self.__updateProgressBar__)
    return


  def repairPropertiesDot(self, projectGroup:str='') -> None:
    """ Repair sqlite database by changing properties table: if key does not have a ., prepend it
    Args:
      projectGroup (str): name of project group
    """
    if not self.projectGroup:
      self.__setBackend__(projectGroup)
    if self.backend is None:
      return
    key= input('Which key to repair, e.g. "chemistry" will become .chemistry? ')
    self.backend.db.cursor.execute(f"SELECT id FROM properties where key == '{key}'")
    res = self.backend.db.cursor.fetchall()
    for idx, docID in enumerate(res):
      docID = docID[0]
      try:
        self.backend.db.cursor.execute(f"UPDATE properties SET key='.{key}' WHERE id == '{docID}' and  key=='{key}'")
      except Exception:
        print(f"Error, could not change {docID} and {key}. Likely that combination exists already in properties. Repair manually")
        if idx==0:
          print(traceback.format_exc())
    self.backend.db.connection.commit()
    print('Done')
    return



def main() -> None:
  '''
  Main function
  '''
  print('\n-------------------------------------------------------------------------')
  print(  '------------     Manual functions for PASTA-ELN installation    ---------')
  print(  '-------------------------------------------------------------------------')
  tools = Tools()
  todos = []
  if len(sys.argv)==2:
    todos = sys.argv[1].split()
    if not re.match(r'^\d+$', todos[0]):
      print('** ERROR Malformed todo list (start with a number), exit')
      return
  tools.run(todos)
  return


if __name__ ==  '__main__':
  main()



"""example addon: adds to the form content.

**This file has to be called form_auto.py and only this file is called from form.**

**In its default form, the description is commented out to disable this addon and not show in the GUI**

THIS IS A ADVANCED ADDON TUTORIAL
This tutorial teaches
- do not use chatgpt library, but use the requests library in order to employ less dependencies
- the structure of documents
"""
import requests

# The following two variables are mandatory
# description  = 'Use Gemini LLM to generate content'  #short description that is shown in the menu
reqParameter = {'key':'API key of google'} #possibility for required parameters: like API-key, etc. {'API': 'text'}
helpText = 'To get API key, you have to register at https://aistudio.google.com/app/apikey and create a new key.'

def main(backend, doc, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        backend (pasta backend): allow to extract data
        doc (dict): dictionary of the content: comment, title, image might be helpful
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        str: content to add
    """
    apiKey = parameter.get('key')
    if not apiKey:
        return 'API key not provided in parameters. Please configure it in the addon settings.'
    if not doc['comment']:
        return 'No comment provided in the document. Please add a comment to expand.'
    promptText = f"Expand the following text: '{doc['comment']}'."
    url        = f'https://generativelanguage.googleapis.com/v1beta/models/gemini-2.0-flash:generateContent?key={apiKey}'
    headers    = {'Content-Type': 'application/json'}
    data       = {'contents': [{'parts': [{'text': promptText}]}]}
    # main try block: if it fails, it will return an error message
    try:
        response = requests.post(url, headers=headers, json=data, timeout=30)
        response.raise_for_status()  # Raises an HTTPError for bad responses (4XX or 5XX)
        result = response.json()
        # text might be result['candidates'][0]['content']['parts'][0]['text']
        reply = result.get('candidates', [{}])[0].get('content', {}).get('parts', [{}])[0].get('text', 'Could not extract text.')
        return reply
    except Exception as e:
        return f"Error calling Gemini API: {e}"



""" Client for communicating with Dataverse Server via REST API
- Author: Jithu Murugan, Steffen Brinckmann
"""
import logging
from datetime import datetime
from json import dumps
from os.path import basename
from typing import Any
from xml.etree.ElementTree import ElementTree, fromstring
# Dataverse: remember to always publish everything before using it!!
import requests
from requests.auth import HTTPBasicAuth
from .dataverseDefaultDict import DATAVERSE_METADATA
from .repository import RepositoryClient


class DataverseClient(RepositoryClient):
  """ Client for communicating with Dataverse Server via REST API """

  def __init__(self, server_url: str, api_token: str, identifier:str) -> None:
    """
    Initializes the client.

    Args:
        server_url (str): The URL of the server.
        api_token (str): The API token for authentication.
        identifier (str): sub-dataverse, category; use '' for void
    """
    super().__init__(server_url, api_token)
    self.identifier = identifier
    self.headers = {'Accept': 'application/json', 'X-Dataverse-key': self.api_token}


  def recreateAPIKey(self) -> str | Any:
    """
    Recreates the API token using existing token.

    Returns:
        str | Any: The new API token or any error message if the token recreation fails.
    """
    resp = requests.post(f"{self.server_url}/api/users/token/recreate", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      token_message = resp.json().get('data').get('message')
      return token_message.replace('New token for dataverseAdmin is ', '')
    return f"Error recreating the token, Info: {resp.text}"


  def checkServer(self) -> tuple[bool, str]:
    """
    Checks if the data-verse server is reachable

    Returns (tuple(bool, Any)):
      A tuple of (success, a message) is returned
    """
    resp = requests.get(f"{self.server_url}/api/info/version", headers={'Accept': 'application/json'}, timeout=10)
    success = (resp.status_code == 200 and resp.json().get('data').get('version') is not None)
    return (success, 'Dataverse is reachable') \
      if success \
      else (success, f"Cannot reach server: {self.server_url}, Status: {resp.status_code}, json: {resp.json()}")


  def checkAPIKey(self) -> bool:
    """
    Checks if the given API token is valid.

    Explanation:
        This method checks if the provided API token is valid by making a request to the server.
        It logs the server URL and sends a GET request to the token endpoint with the API token.
        It returns True if the response is successful and the status code is not 401, 403, or 500.

    Args:
        self: The instance of the class.

    Returns:
        bool: True if the API token is valid, False otherwise.
    """
    resp = requests.get(f"{self.server_url}/api/users/token", headers=self.headers, timeout=10)
    return (bool(resp) and resp.status_code is not None
            and resp.status_code not in [401, 403, 500])


  def checkAPIKeyExpired(self) -> bool:
    """
    Check if the API token has expired
    Returns:
      True if the token has expired, False otherwise
    """
    resp = requests.get(f"{self.server_url}/api/users/token", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      expiry_message = resp.json().get('data').get('message')
      expiry_time_string = expiry_message.replace(f"Token {self.api_token} expires on ", '')
      expiry_time = datetime.strptime(expiry_time_string, '%Y-%m-%d %H:%M:%S.%f')
      return expiry_time < datetime.now()
    print(f"Error checking token expiration, Info: {resp.text}")
    return False


  ################ DATAVERSE #######################
  def createDataverse(self, dv_parent: str, dv_name: str, dv_description: str, dv_alias: str,
                      dv_contact_email_list: list[dict[str, str]], dv_affiliation: str, dv_type: str) -> dict[Any, Any] | Any:
    """
    Creates and publishes a dataverse
    Args:
      dv_parent (str): The name of the parent dataverse
      dv_name (str): The name of the dataverse
      dv_description (str): The description of the dataverse
      dv_alias (str): The alias of the dataverse
      dv_contact_email_list (list[dict[str, str]]): The list of contact emails which must be provided in the following format:
        [{"contactEmail": "example1@example.edu"}, {"contactEmail": "example2@example.edu"}]
      dv_affiliation (str): The affiliation of the dataverse
      dv_type (str): The type of the dataverse which must be one of the following:
        DEPARTMENT
        JOURNALS
        LABORATORY
        ORGANIZATIONS_INSTITUTIONS
        RESEARCHERS
        RESEARCH_GROUP
        RESEARCH_PROJECTS
        TEACHING_COURSES
        UNCATEGORIZED

    Returns(dict[Any, Any] | Any):
      A dictionary representing the created and published dataverse, or an error message.
    """
    dv_json = {
      'name': dv_name,
      'alias': dv_alias,
      'dataverseContacts': dv_contact_email_list,
      'affiliation': dv_affiliation,
      'description': dv_description,
      'dataverseType': dv_type
    }
    # Create the data-verse
    resp = requests.post(f"{self.server_url}/api/dataverses/{dv_parent}",
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token},
      data=dumps(dv_json), timeout=10)
    if resp.status_code == 201:  # Success
      pub_resp = requests.post(f"{self.server_url}/api/dataverses/{resp.json().get('data').get('alias')}/actions/:publish",
        headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, timeout=10)
      if pub_resp.status_code == 200:
        return pub_resp.json().get('data')
      return f"Error publishing dataverse, Status: {pub_resp.status_code}, Info: {pub_resp.text}"
    return f"Error creating dataverse, Status: {resp.status_code}, Info: {resp.text}"


  def getDataverseList(self) -> dict[Any, Any] | Any:
    """
    Gets the list of data verses
    Returns:
      A dictionary of dataverses (identifier & title) for successful request,
      otherwise the error message is returned.
    """
    resp = requests.get(f"{self.server_url}/dvn/api/data-deposit/v1.1/swordv2/service-document",
      headers={'Accept': 'application/json', 'X-Dataverse-key': self.api_token},
      auth=HTTPBasicAuth(self.api_token, ''), timeout=10)
    if resp.status_code == 200:
      element_tree: ElementTree = ElementTree(fromstring(resp.text))
      dataverse_list: list[dict[str, str]] = []
      for element in element_tree.getroot().findall('.//{http://www.w3.org/2007/app}collection'):# type:ignore[union-attr]
        title = element.find('.//{http://www.w3.org/2005/Atom}title')
        if title is not None:
          title_val = title.text if title.text is not None else ''
          dataverse_list.append({'id': element.attrib['href'].split('/')[-1],'title': title_val})
      dataverse_list.sort(key=lambda x: x['title'])
      return dataverse_list
    return f"Error get dataverse list, Server:{self.server_url},  Status:{resp.status_code}"


  def getDataverseContent(self) -> dict[Any, Any] | Any:
    """
    Retrieves the contents of a dataverse.

    Returns (dict[Any, Any] | Any):
      A dictionary of dataverse contents for successful request, otherwise the error message is returned.
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}/contents", headers=self.headers,
                        timeout=10)
    return resp.json() if resp.status_code == 200 else \
      f"Error retrieving the contents of dataverse, Id: {self.identifier}, Info: {resp.json()}"


  def getDataverseSize(self) -> str | Any:
    """
    Retrieves the size of a dataverse.
    Returns (str):
      Dataverse size in bytes for successful request, otherwise the error message is returned.
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}/storagesize", headers=self.headers,
                        timeout=10)
    if resp.status_code == 200:
      return (resp.json().get('data').get('message').replace('Total size of the files stored in this dataverse: ', ''))
    return f"Error retrieving the size for data verse, Id: {self.identifier}, Info: {resp.json()}"


  def getDataverseInfo(self) -> str | Any:
    """
    Retrieves information about a dataverse.

    Returns:
        str | Any: The data associated with the dataverse if the request is successful, otherwise an error message.
    """
    resp = requests.get(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers,
                        timeout=10)
    if resp.status_code:
      return resp.json().get('data')
    return f"Error retrieving the info for data verse, Id: {self.identifier}, Info: {resp.text}"


  ################ DATASET #######################
  def createDataset(self, ds_metadata: dict[str, Any], ds_validate_metadata: bool = False
                                       ) -> dict[Any, Any] | Any:
    """
    Creates and publishes a dataset to the parent dataverse.
    Args:
      ds_metadata (dict[str, Any]): The dataset metadata.
        Refer the https://guides.dataverse.org/en/latest/_downloads/4e04c8120d51efab20e480c6427f139c/dataset-create-new-all-default-fields.json for the default values to be used in the metadata.
        The type names to be used in the metadata along with the values can be found in the metadata blocks and should correspond to the dataset-create-new-all-default-fields.json.
      ds_validate_metadata (bool): Whether to validate the metadata.

    Returns:
      A dictionary of dataset metadata with the persistent identifier for successful request, otherwise the error message is returned.
    """
    metadata = DATAVERSE_METADATA
    if 'license' in ds_metadata:
      metadata['datasetVersion']['license'] = ds_metadata['license']
    else:
      del metadata['datasetVersion']['license']
    for _, metablock in metadata['datasetVersion']['metadataBlocks'].items():  # type:ignore[attr-defined]
      field_copy = metablock['fields'].copy()
      del metablock['displayName']
      metablock['fields'].clear()
      for field in field_copy:
        if field['typeName'] in ds_metadata:
          field['value'] = ds_metadata[field['typeName']]
          metablock['fields'].append(field)
    # Request to create the dataset
    resp = requests.post(f"{self.server_url}/api/dataverses/{self.identifier}/datasets",
      params={'doNotValidate': str(not ds_validate_metadata)}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, json=metadata)
    if resp.status_code == 201:
      # Request to publish the dataset
      resp = requests.post(f"{self.server_url}/api/datasets/:persistentId/actions/:publish",
        params={'persistentId': resp.json().get('data').get('persistentId'), 'type': 'major'},
        headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token}, timeout=10)
      if resp.status_code == 200:
        return resp.json().get('data')
      return f"Error publishing dataset, Info: {resp.text}"
    return f"Error creating dataset, Info: {resp.text}"


  def uploadFile(self, ds_pid: str, df_file_path: str, df_description: str, df_categories: list[str]) -> dict[Any, Any] | Any:
    """
    Uploads a file to a dataset.
    Args:
      ds_pid (str): The identifier of the dataset.
      df_file_path (str): The absolute path to the file to be uploaded.
      df_description (str): The description of the file.
      df_categories (list[str]): The categories/tags for the file.

    Returns:
      {   'file_upload_result': file_upload_response,
          'dataset_publish_result': dataset_publish_response
      } for successful request, otherwise the error message is returned.
    """
    filename = basename(df_file_path)
    metadata = dumps({'description': df_description, 'categories': df_categories})
    data:Any = {}
    with open(df_file_path, 'rb') as file_stream:
      data['file'] = (filename, file_stream, 'multipart/form-data')
      data['jsonData'] = (None, metadata, 'application/json')
      # Request to add the file to dataset
      resp = requests.post(
        f"{self.server_url}/api/datasets/:persistentId/add",
        params={'persistentId': ds_pid},
        headers={'X-Dataverse-key': self.api_token},
        files=data,
        timeout=5)
      if resp.status_code == 200:
        # Request to publish the dataset
        pub_resp = requests.post(
          f"{self.server_url}/api/datasets/:persistentId/actions/:publish",
          params={'persistentId': ds_pid, 'type': 'major'}, timeout=10,
          headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
        if pub_resp.status_code == 200:
          return {'file_upload_result': resp.json().get('data'),
                  'dataset_publish_result': pub_resp.json().get('data')}
        return f"Error publishing dataset: {ds_pid} as part of file ({df_file_path}) upload on server: "\
               f"{self.server_url}, Info: {pub_resp.json()}"
      return f"Error uploading file: {df_file_path} to dataset: {ds_pid} Info: {resp.json()}"


  def uploadRepository(self, metadata:dict[str,Any], file_path:str) -> tuple[bool, str]:
    """
    Uploads a file and metadata to become a dataset.

    Args:
      metadata (dict): metadata to this file according to dataverse standard
      file_path (str): The absolute path to the file to be uploaded.

    Returns:
      tuple: success of function, message
    """
    res= self.createDataset(metadata)
    if isinstance(res, str):
      return False, f'Error publishing the dataset: {res}'
    doi = f"{res['protocol']}:{res['authority']}/{res['identifier']}"
    reply = self.uploadFile(doi, file_path, '.eln file', ['file'])
    if isinstance(reply, str):
      return False, 'Error publishing the file'
    return True, f'Published: {doi}, {res["persistentUrl"]}'


  def getDatasetInfo(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch JSON representation of a dataset.
    Args:
      ds_persistent_id (str): The identifier of the dataset.
      version (str): The version of the dataset.
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version.
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number.
          x same as x.0

    Returns:
      JSON representation of the dataset for successful request, otherwise the error message is returned.
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching JSON representation of dataset: {ds_persistent_id} Info: {resp.json()}"


  def getDatasetVersions(self, ds_persistent_id: str) -> dict[Any, Any] | Any:
    """
    Fetch the version list for dataset.
    Args:
      ds_persistent_id (str): The identifier of the dataset.

    Returns:
      Version list for the dataset for successful request, otherwise the error message is returned.
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching version list for dataset: {ds_persistent_id} Info: {resp.json()}"


  def getDatasetLocks(self, ds_persistent_id: str) -> dict[Any, Any] | Any:
    """
    Fetches locks for a dataset.

    Explanation:
        This method fetches the locks for a dataset identified by the persistent ID.
        It makes an asynchronous HTTP GET request to retrieve the locks information and handles exceptions.

    Args:
        ds_persistent_id (str): The persistent ID of the dataset to fetch locks for.

    Returns:
        dict[Any, Any] | Any: A dictionary containing the locks information if successful, or an error message.
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/locks?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return {'locks': resp.json().get('data')}
    return f"Error fetching locks for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def getDatasetFiles(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch the file list for dataset.
    Args:
      ds_persistent_id (str): The identifier of the dataset.
      version (str): The version of the dataset.
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version.
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number.
          x same as x.0

    Returns:
      File list for the dataset for successful request, otherwise the error message is returned.
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}/files?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching file list for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def getDatasetMetadata(self, ds_persistent_id: str, version: str = ':latest-published') -> dict[Any, Any] | Any:
    """
    Fetch the metadata block for dataset.
    Args:
      ds_persistent_id (str): The identifier of the dataset.
      version (str): The version of the dataset.
        Note dataset versions can be referred to as:
          :draft the draft version, if any
          :latest either a draft (if exists) or the latest published version.
          :latest-published the latest published version
          x.y a specific version, where x is the major version number and y is the minor version number.
          x same as x.0

    Returns:
      Metadata block for the dataset for successful request, otherwise the error message is returned.
    """
    resp = requests.get(
      f"{self.server_url}/api/datasets/:persistentId/versions/{version}/metadata?persistentId={ds_persistent_id}",
      params={'Accept': 'application/json'}, timeout=10,
      headers={'Content-Type': 'application/json', 'X-Dataverse-key': self.api_token})
    if resp.status_code == 200:
      return resp.json().get('data')
    return f"Error fetching metadata block for dataset: {ds_persistent_id}  Info: {resp.json()}"


  def deleteEmptyDataverse(self) -> str | Any:
    """
    Deletes an empty dataverse.

    Returns:
      Message for successful request, otherwise the error message is returned.
    """
    resp = requests.delete(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataverse,  Info: {resp.json()}"


  def deletePublishedDataset(self, ds_persistent_id: str) -> str | Any:
    """
    Deletes a published dataset.
    Note: The method fails presently due to a bug in dataverse API.
          Every time the API returns 403 errors saying the user needs to be superuser.
    Args:
      ds_persistent_id (str): The identifier of the dataset.

    Returns:
      Message for successful request, otherwise the error message is returned.
    """
    resp = requests.delete(
      f"{self.server_url}/api/datasets/:persistentId/destroy/",
      params={'persistentId': ds_persistent_id},
      headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataset, Id: {ds_persistent_id}, Info: {resp.json()}"


  def deleteNonEmptyDataverse(self) -> str | Any:
    """
    Deletes a non-empty dataverse.

    Note: The method fails presently due to a bug in dataverse API while invoking the delete_published_dataset method.

    Returns:
      Message for successful request, otherwise the error message is returned.
    """
    contents = self.getDataverseContent()
    for content in contents:
      if content.get('type') == 'dataset':
        self.deletePublishedDataset(f"{content.get('protocol')}:{content.get('authority')}/{content.get('identifier')}")
      elif content.get('type') == 'dataverse':
        self.deleteNonEmptyDataverse()
      else:
        logging.error('Unknown content type: %s while deleting dataverse: %s on server: %s ', content.type,
                      self.identifier, self.server_url)
    resp = requests.delete(f"{self.server_url}/api/dataverses/{self.identifier}", headers=self.headers, timeout=10)
    if resp.status_code == 200:
      return resp.json().get('data').get('message')
    return f"Error deleting dataverse, Id: {self.identifier}, Info: {resp.json()}"


  def prepareMetadata(self, metadata:dict[str,Any]) -> dict[str,Any]:
    """
    Prepares the metadata for uploading.

    Args:
        metadata (dict): The metadata to be prepared.

    Returns:
        dict: The prepared metadata.
    """
    author = metadata['author']
    fields = [{'typeName': 'title', 'value': metadata['title'], 'typeClass': 'primitive'},
              {'typeName': 'author', 'value': [{'authorName': {'value': f"{author['last']}, {author['first']}"},
                'authorIdentifier': {'value': author['orcid']},
                'authorAffiliation': {'value': author['organizations'][0]['organization']}}], 'typeClass': 'compound'},
              {'typeName': 'datasetContact', 'value': [{'datasetContactEmail': {'value': author['email']},
                'datasetContactName': {'value': f"{author['last']}, {author['first']}"}}], 'typeClass': 'compound'},
              {'typeName': 'keywords', 'value': metadata['keywords'], 'typeClass': 'primitive'},
              {'typeName': 'publicationDate', 'value': datetime.now().strftime('%Y-%m-%d'), 'typeClass': 'primitive'},
              {'typeName': 'dsDescription', 'value': [{'dsDescriptionValue': {'value': metadata['description']}}],
                'typeClass': 'compound'},
              {'typeName': 'subject', 'value': [metadata['category']], 'typeClass': 'controlledVocabulary'}
            ] + metadata['additional']
    return {'metadata': {'datasetVersion': {'metadataBlocks': {'citation': {'fields': fields}}}}}



"""example addon: plot data from a table

THIS IS A VERY ADVANCED ADDON TUTORIAL
This tutorial teaches
- how to plot the metadata from a table
"""
import matplotlib.pyplot as plt
import pandas as pd
from pasta_eln.miscTools import dfConvertColumns

# The following two variables are mandatory
description  = 'Default metadata plot'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}

def main(backend, df, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        backend (pasta backend): allow to extract data
        df (Dataframe): dataframe with the data to plot
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        bool: success
    """
    df = dfConvertColumns(df, 10) # every 10th value should be numeric to convert to numeric
    for col in df.columns:
        numeric_series = pd.to_numeric(df[col], errors='coerce')
        if numeric_series.notna().any():
            plt.plot(df[col],'o')
            plt.xlabel('Index')
            plt.ylabel(col)
            plt.show()
            break
    return True



""" Terminology Lookup Dialog class which handles the IRI lookup online """
#  PASTA-ELN and all its sub-parts are covered by the MIT license.
#
#  Copyright (c) 2023
#
#  Author: Jithu Murugan
#  Filename: terminology_lookup_dialog.py
#
#  You should have received a copy of the license with this file. Please refer the license file for more information.
import textwrap
from asyncio import get_event_loop
from typing import Callable
from PySide6 import QtCore, QtWidgets
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QCheckBox, QHBoxLayout, QLabel, QMessageBox, QWidget
from pasta_eln.GUI.definitions.dialog_base import Ui_TerminologyLookupDialogBase  # type: ignore[attr-defined]
from pasta_eln.GUI.definitions.terminology_lookup_service import TerminologyLookupService


class TerminologyLookupDialog(Ui_TerminologyLookupDialogBase):
  """ Terminology Lookup Dialog class which handles the IRI lookup online """

  def __init__(self,
               default_lookup_term: str | None = None,
               accepted_callback: Callable[[list[str]], None] = None) -> None:  # type: ignore[assignment]
    """
    Initializes the dialog
    Args:
      default_lookup_term (str): Default search term to be used by the terminology lookup service.
      accepted_callback (Callable[[], None]): Accepted button parent callback.
    """
    self.accepted_callback: Callable[[list[str]], None] = accepted_callback
    # Set up the UI elements
    self.instance = QtWidgets.QDialog()
    super().setupUi(self.instance)
    self.terminology_lookup_service = TerminologyLookupService()
    # User selected urls
    self.selected_iris: list[str] = []
    # Load the icon images for lookup portals
    self.icons_pixmap = self.terminology_lookup_service.getIconDict()
    # Hide the error console and connect the slot
    self.errorConsole.hide()
    self.errorConsoleBtn.clicked.connect(lambda: self.errorConsole.setVisible(not self.errorConsole.isVisible()))
    self.buttonBox.accepted.connect(self.set_selected_iris)
    self.buttonBox.accepted.connect(lambda: self.accepted_callback(self.selected_iris))
    self.terminologySearchPushButton.clicked.connect(self.terminology_search_button_clicked)
    self.terminologyLineEdit.setText(default_lookup_term)
    self.terminology_search_button_clicked()


  def show(self) -> None:
    """ Displays the dialog """
    self.instance.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
    self.instance.show()
    return


  def add_scroll_area_entry(self, pixmap: QPixmap, checkbox_text: str, checkbox_tooltip: str) -> None:
    """
    Adds an entry to the scroll area of terminology lookup dialog.
    Entry consists of a checkbox and a label
    The search result is added as a checkbox and a label (icon) to the end of the entry
    Args:
      pixmap (QPixmap): Icon image to be added
      checkbox_text (str): Text to be added to the checkbox
      checkbox_tooltip (str): Tooltip for the checkbox
    """
    # Set up the layout for the entry with check box and label
    entry_layout = QHBoxLayout()
    check_box = QCheckBox(checkbox_text)
    check_box.setToolTip(checkbox_tooltip)
    entry_layout.addWidget(check_box)
    entry_layout.addStretch(1)
    entry_layout.addWidget(QLabel(pixmap=pixmap))  # type: ignore[call-overload]
    # Create a widget for the entry with the created layout
    entry_widget = QWidget()
    entry_widget.setLayout(entry_layout)
    self.scrollAreaContentsVerticalLayout.addWidget(entry_widget)
    return


  def clear_scroll_area(self) -> None:
    """ Clears the scroll area by removing all the widgets """
    for widget_pos in reversed(range(self.scrollAreaContentsVerticalLayout.count())):
      self.scrollAreaContentsVerticalLayout.itemAt(widget_pos).widget().setParent(None)
    return


  def set_selected_iris(self) -> None:
    """
    Gets the IRIs from the checked QCheckBoxes of the scroll area and appends them to the list of selected IRIs
    Tooltip of the checked QCheckBox holds the IRI information
    """
    self.selected_iris.clear()
    for widget_pos in range(self.scrollAreaContentsVerticalLayout.count()):
      check_box = self.scrollAreaContentsVerticalLayout.itemAt(widget_pos).widget().findChildren(QCheckBox)[0]
      if check_box.isChecked():
        self.selected_iris.append(check_box.toolTip())
    return


  def terminology_search_button_clicked(self) -> None:
    """
    terminologySearchPushButton Button click event handler which initiates
    the terminology search, retrieve the results and updated the scroll area
    """
    self.reset_ui()
    search_term = self.terminologyLineEdit.text()
    if not search_term or search_term.isspace():
      QMessageBox.warning(self.instance, 'Error', 'Enter non null search term!',
                          QMessageBox.StandardButton.NoButton, QMessageBox.StandardButton.Ok)
      return
    self.searchProgressBar.setValue(5)
    event_loop = get_event_loop()
    if lookup_results := event_loop.run_until_complete(
        self.terminology_lookup_service.do_lookup(search_term)):
      for service in lookup_results:
        for result in service['results']:
          self.add_scroll_area_entry(self.icons_pixmap[service['name']],
                                     textwrap.fill(result['information'], width=100, max_lines=2),
                                     result['iri'])
          self.searchProgressBar.setValue((100 - self.searchProgressBar.value()) / 2)
    self.searchProgressBar.setValue(100)
    return


  def reset_ui(self) -> None:
    """ Resets the UI elements for the dialog """
    self.searchProgressBar.setValue(0)
    self.clear_scroll_area()
    self.errorConsole.clear()
    self.errorConsole.setVisible(False)
    self.selected_iris.clear()
    return


if __name__ == '__main__':
  import sys
  app = QtWidgets.QApplication(sys.argv)
  ui = TerminologyLookupDialog('Project')
  ui.instance.show()
  sys.exit(app.exec())



""" Main class of config tab on parameters (e.g. API keys) for add-ons """
import importlib
import json
from pathlib import Path
from typing import Callable
from PySide6.QtWidgets import QDialog, QDialogButtonBox, QLineEdit, QVBoxLayout  # pylint: disable=no-name-in-module
from ..fixedStringsJson import CONF_FILE_NAME
from ..guiCommunicate import Communicate
from ..guiStyle import Label, TextButton, showMessage, widgetAndLayout


class ConfigurationAddOnParameter(QDialog):
  """ Main class of config tab on authors
  """
  def __init__(self, comm:Communicate, callbackFinished:Callable[[bool],None]):
    """
    Initialization

    Args:
      comm (Communicate): communication
      callbackFinished (function): callback function to call upon end
    """
    super().__init__()
    self.comm = comm
    self.callbackFinished = callbackFinished
    mainL = QVBoxLayout(self)
    Label('Define Add-On parameters','h2',mainL)

    #GUI elements
    self.allLineEdits = []
    if hasattr(comm.backend, 'configuration'):
      addOns = comm.backend.configuration['projectGroups'][comm.backend.configurationProjectGroup]['addOns']
      for addOnType in addOns:                               # loop over add-on types
        if addOnType != 'extractors' and addOns[addOnType]:
          for name, _ in addOns[addOnType].items():          # loop over add-ons
            module      = importlib.import_module(name)
            requiredParam = module.reqParameter
            try:
              helpText = module.helpText
            except AttributeError:
              helpText = ''
            for param, tooltip in requiredParam.items():    # loop over parameters
              _, barL = widgetAndLayout('H', mainL, 'm', 's', 's', 's', 's')
              Label(f'{addOnType}/{name}.py: {param}', 'h4', barL, tooltip=tooltip)
              lineEdit = QLineEdit()
              barL.addWidget(lineEdit)
              if helpText:
                TextButton('?', self, command=[helpText], layout=barL)
              self.allLineEdits.append((addOnType,name,param,lineEdit))

    #final button box
    mainL.addStretch(1)
    buttonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Save | QDialogButtonBox.StandardButton.Cancel)
    buttonBox.clicked.connect(self.closeDialog)
    mainL.addWidget(buttonBox)


  def execute(self, command:list[str]) -> None:
    """
    Execute a command

    Args:
      command (list[str]): command to execute
    """
    showMessage(self, 'Help', command[0], 'Information')
    return


  def closeDialog(self, btn:TextButton) -> None:
    """
    cancel or save entered data

    Args:
      btn (QButton): save or cancel button
    """
    if btn.text().endswith('Cancel'):
      self.reject()
    else:
      apiKeys = self.comm.backend.configuration.get('addOnParameter',{})
      for _, name, param, lineEdit in self.allLineEdits:
        if name not in apiKeys:
          apiKeys[name] = {}
        apiKeys[name][param] = lineEdit.text()
      self.comm.backend.configuration['addOnParameter'] = apiKeys
      with open(Path.home()/CONF_FILE_NAME, 'w', encoding='utf-8') as fConf:
        fConf.write(json.dumps(self.comm.backend.configuration,indent=2))
      self.accept()
    self.callbackFinished(False)
    return



""" Entire config dialog (dialog is blocking the main-window, as opposed to create a new widget-window)"""
from PySide6.QtWidgets import QDialog, QTabWidget, QVBoxLayout  # pylint: disable=no-name-in-module
from ..guiCommunicate import Communicate
from .configAddOnParameter import ConfigurationAddOnParameter
from .configAuthors import ConfigurationAuthors
from .configGUI import ConfigurationGUI
from .configProjectGroup import ProjectGroup
from .configSetup import ConfigurationSetup
from .repositories.configGUI import ConfigurationRepositories


class Configuration(QDialog):
  """ Main class of entire config dialog """
  def __init__(self, comm:Communicate, startTab:str=''):
    """
    Initialization
    - if sidebar.py notices no database, it will call this dialog with 'setup' as only tab

    Args:
      backend (Pasta): backend, not communication
      startTab (str): tab to show initially
    """
    super().__init__()
    self.comm = comm
    self.setWindowTitle('PASTA-ELN configuration')
    self.setMinimumWidth(1200)

    # GUI elements
    mainL = QVBoxLayout(self)
    tabW = QTabWidget(self)
    mainL.addWidget(tabW)

    # tab has to always exist
    tabSetup = ConfigurationSetup(self.comm, self.closeWidget)      # Setup / Troubleshoot Pasta

    # optional tabs
    try:
      tabProjectGroup = ProjectGroup(self.comm, self.closeWidget)     # Project group. Restart app
      tabW.addTab(tabProjectGroup, 'Project group')

      tabGUI = ConfigurationGUI(self.comm, self.closeWidget)          # Misc configuration: e.g. theming. Restart app
      tabW.addTab(tabGUI, 'Appearance')

      tabAuthors = ConfigurationAuthors(self.comm, self.closeWidget)  # Author(s)
      tabW.addTab(tabAuthors, 'Author')

      tabRepository = ConfigurationRepositories(self.comm, self.closeWidget)  # Repositories
      tabW.addTab(tabRepository, 'Repository')

      tabAddOnParameter = ConfigurationAddOnParameter(self.comm, self.closeWidget)  # Add-on parameters
      tabW.addTab(tabAddOnParameter, 'Add-on parameters')

      # initialize when setup is called
      if startTab=='setup':
        tabW.setCurrentWidget(tabSetup)
        tabW.setTabEnabled(0, False)
        tabW.setTabEnabled(1, False)
        tabW.setTabEnabled(2, False)
        tabW.setTabEnabled(3, False)
        tabW.setTabEnabled(4, False)
    except Exception as e:
      print('**ERROR: could not create configuration dialog:', e)

    # always add setup tab
    tabW.addTab(tabSetup, 'Setup')


  def closeWidget(self, restart:bool=True) -> None:
    """
    callback function to close widget, without restarting the entire app

    Args:
      restart (bool): if true, initialize
    """
    self.close()
    if restart:
      self.comm.restart.emit()        #send signal
    return



# type: ignore
# Form implementation generated from reading ui file 'terminology_lookup_dialog_base.ui'
#
# Created by: PyQt6 UI code generator 6.4.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.
from PySide6 import QtCore, QtWidgets


class Ui_TerminologyLookupDialogBase:
  def setupUi(self, TerminologyLookupDialogBase):
    TerminologyLookupDialogBase.setObjectName('TerminologyLookupDialogBase')
    TerminologyLookupDialogBase.resize(900, 767)
    self.gridLayout = QtWidgets.QGridLayout(TerminologyLookupDialogBase)
    self.gridLayout.setObjectName('gridLayout')
    self.buttonBox = QtWidgets.QDialogButtonBox(parent=TerminologyLookupDialogBase)
    self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
    self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
    self.buttonBox.setObjectName('buttonBox')
    self.gridLayout.addWidget(self.buttonBox, 2, 0, 1, 1)
    self.mainVerticalLayout = QtWidgets.QVBoxLayout()
    self.mainVerticalLayout.setObjectName('mainVerticalLayout')
    self.searchBarHorizontalLayout = QtWidgets.QHBoxLayout()
    self.searchBarHorizontalLayout.setObjectName('searchBarHorizontalLayout')
    self.terminologyLineEdit = QtWidgets.QLineEdit(parent=TerminologyLookupDialogBase)
    self.terminologyLineEdit.setClearButtonEnabled(True)
    self.terminologyLineEdit.setObjectName('terminologyLineEdit')
    self.searchBarHorizontalLayout.addWidget(self.terminologyLineEdit)
    self.terminologySearchPushButton = QtWidgets.QPushButton(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.terminologySearchPushButton.sizePolicy().hasHeightForWidth())
    self.terminologySearchPushButton.setSizePolicy(sizePolicy)
    self.terminologySearchPushButton.setMinimumSize(QtCore.QSize(150, 0))
    self.terminologySearchPushButton.setObjectName('terminologySearchPushButton')
    self.searchBarHorizontalLayout.addWidget(self.terminologySearchPushButton)
    self.mainVerticalLayout.addLayout(self.searchBarHorizontalLayout)
    self.searchProgressBar = QtWidgets.QProgressBar(parent=TerminologyLookupDialogBase)
    self.searchProgressBar.setProperty('value', 0)
    self.searchProgressBar.setObjectName('searchProgressBar')
    self.mainVerticalLayout.addWidget(self.searchProgressBar)
    self.scrollAreaHorizontalLayout = QtWidgets.QHBoxLayout()
    self.scrollAreaHorizontalLayout.setObjectName('scrollAreaHorizontalLayout')
    self.scrollArea = QtWidgets.QScrollArea(parent=TerminologyLookupDialogBase)
    self.scrollArea.setWidgetResizable(True)
    self.scrollArea.setObjectName('scrollArea')
    self.scrollAreaWidgetContents = QtWidgets.QWidget()
    self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 876, 415))
    self.scrollAreaWidgetContents.setObjectName('scrollAreaWidgetContents')
    self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
    self.gridLayout_2.setObjectName('gridLayout_2')
    self.scrollAreaContentsVerticalLayout = QtWidgets.QVBoxLayout()
    self.scrollAreaContentsVerticalLayout.setObjectName('scrollAreaContentsVerticalLayout')
    self.gridLayout_2.addLayout(self.scrollAreaContentsVerticalLayout, 0, 0, 1, 1)
    self.scrollArea.setWidget(self.scrollAreaWidgetContents)
    self.scrollAreaHorizontalLayout.addWidget(self.scrollArea)
    self.mainVerticalLayout.addLayout(self.scrollAreaHorizontalLayout)
    self.errorConsoleVerticalLayout = QtWidgets.QVBoxLayout()
    self.errorConsoleVerticalLayout.setObjectName('errorConsoleVerticalLayout')
    self.errorButtonsHorizontalLayout = QtWidgets.QHBoxLayout()
    self.errorButtonsHorizontalLayout.setObjectName('errorButtonsHorizontalLayout')
    self.errorConsoleBtn = QtWidgets.QPushButton(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Fixed, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.errorConsoleBtn.sizePolicy().hasHeightForWidth())
    self.errorConsoleBtn.setSizePolicy(sizePolicy)
    self.errorConsoleBtn.setMinimumSize(QtCore.QSize(150, 0))
    self.errorConsoleBtn.setObjectName('errorConsoleBtn')
    self.errorButtonsHorizontalLayout.addWidget(self.errorConsoleBtn)
    spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum)
    self.errorButtonsHorizontalLayout.addItem(spacerItem)
    self.errorConsoleVerticalLayout.addLayout(self.errorButtonsHorizontalLayout)
    self.errorConsole = QtWidgets.QTextEdit(parent=TerminologyLookupDialogBase)
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Preferred, QtWidgets.QSizePolicy.Policy.Fixed)
    sizePolicy.setHorizontalStretch(0)
    sizePolicy.setVerticalStretch(0)
    sizePolicy.setHeightForWidth(self.errorConsole.sizePolicy().hasHeightForWidth())
    self.errorConsole.setSizePolicy(sizePolicy)
    self.errorConsole.setMinimumSize(QtCore.QSize(0, 0))
    self.errorConsole.setObjectName('errorConsole')
    self.errorConsoleVerticalLayout.addWidget(self.errorConsole)
    self.mainVerticalLayout.addLayout(self.errorConsoleVerticalLayout)
    self.gridLayout.addLayout(self.mainVerticalLayout, 0, 0, 1, 1)

    self.retranslateUi(TerminologyLookupDialogBase)
    self.buttonBox.accepted.connect(TerminologyLookupDialogBase.accept) # type: ignore
    self.buttonBox.rejected.connect(TerminologyLookupDialogBase.reject) # type: ignore
    QtCore.QMetaObject.connectSlotsByName(TerminologyLookupDialogBase)

  def retranslateUi(self, TerminologyLookupDialogBase):
    _translate = QtCore.QCoreApplication.translate
    TerminologyLookupDialogBase.setWindowTitle(_translate('TerminologyLookupDialogBase', 'Terminology Lookup'))
    self.terminologyLineEdit.setPlaceholderText(_translate('TerminologyLookupDialogBase', "\"Search for Definitions in Wikis/Ontologies\""))
    self.terminologySearchPushButton.setText(_translate('TerminologyLookupDialogBase', 'Search'))
    self.errorConsoleBtn.setText(_translate('TerminologyLookupDialogBase', 'Show/Hide Errors'))


if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication(sys.argv)
    TerminologyLookupDialogBase = QtWidgets.QDialog()
    ui = Ui_TerminologyLookupDialogBase()
    ui.setupUi(TerminologyLookupDialogBase)
    TerminologyLookupDialogBase.show()
    sys.exit(app.exec())



"""example addon: fill definitions table with first entry from wikidata

**This file has to be called definition_autofill.py and only this file is called from the definitions GUI.**

**In its default form, the description is commented out to disable this addon and not show in the GUI**

THIS IS A ADVANCED ADDON TUTORIAL
This tutorial teaches
- the structure of the definitions table
- how to fill the definitions table with the first entry from wikidata
- these addons can have multiple functions inside them
"""
import requests

# The following two variables are mandatory
# description  = 'Fill the definitions table by the first entry from wikidata'  #short description that is shown in the menu
reqParameter = {} #possibility for required parameters: like API-key, etc. {'API': 'text'}


def getFirstWikidataEntry(definition):
    """Get the first Wikidata entry for a given definition
    Args:
        definition (str): The definition to search for
    Returns:
        str: The URL of the first Wikidata entry, or '' if not found
    """
    url = f'https://www.wikidata.org/w/api.php?action=wbsearchentities&search={definition}&language=en&format=json'
    response = requests.get(url)
    if response.status_code == 200:  # success
        data = response.json()
        if 'search' in data and len(data['search']) > 0: # check if search results exist
            return data['search'][0]['concepturi']
    return ''


def main(backend, df, widget, parameter={}):
    """ main function: has to exist and is called by the menu
    Args:
        backend (pasta backend): allow to extract data
        df (DataFrame): pandas dataframe with data
           Columns are: 'key', 'long', 'PURL', 'defType' with the PURL being filled
        widget (QWidget): allows to create new gui dialogs
        parameter (dict): ability to pass parameters

    Returns:
        Dataframe: modified dataframe
    """
    if df.empty:
        return df
    emptyPURLmask = df['PURL'].isnull() | (df['PURL'] == '')  # only fill empty PURLs
    df.loc[emptyPURLmask, 'PURL'] = df.loc[emptyPURLmask, 'key'].apply(getFirstWikidataEntry)
    return df



#!/usr/bin/env python
""" setup to generate version number"""
from __future__ import annotations

from setuptools import setup

import releaseVersion

if __name__ >= '__main__':
  setup(name='pasta_eln', version=releaseVersion.getVersion()[1:])
